from enum import Enum

from PySide6.QtCore import QFile, QSize, Qt
from PySide6.QtGui import QCursor, QIcon, QPainter, QPixmap
from PySide6.QtSvg import QSvgRenderer
from PySide6.QtWidgets import QPushButton

from houdini_package_manager.meta.meta_tools import StatusBar
from houdini_package_manager.utils import epath


class BtnIcon(Enum):
    """
    Enum class for different types of SVG QPushButton icons.
    Each string is the filename of the respective SVG file.
    """

    SOURCE_CONTROL = "source_control"
    FILE = "file"
    UPDATE = "update"
    HPM_LOGO = "hpm"
    HPM_LOGO_GREY = "hpm_grey"
    ADD_PKG = "add_packages"
    ADD_PKG_CONFIRM = "add_packages_confirm"
    REPO = "repo"
    BUG = "bug"
    DOCS = "docs"
    FOLDER = "folder"
    GIT_SYNC = "git_sync"
    MIGRATE = "migrate"
    REFRESH = "refresh"
    REFRESH_ALL = "refresh_all"
    WARNING = "warning"

    ADD = "add"
    REMOVE = "remove"
    REMOVE_ALL_ITEMS = "remove_all_items"
    INVERT_SELECTION = "invert_selection"
    CLEAR_SELECTION = "clear_selection"


class BtnSize(Enum):
    """
    Enum class for different types of SVG QPushButton pixel sizes.
    """

    SQUARE_DEFAULT = (28, 28)
    SQUARE_MEDIUM = (24, 24)
    SQUARE_SMALL = (16, 16)
    WIDE_DEFAULT = (48, 24)
    WIDE_LARGE = (120, 36)

    CELL_TALL = (23, 29)

    """SPECIALS"""
    HPM_LOGO = (56, 28)
    FOLDER = (38, 24)
    WARNING = (32, 29)
    DOCS = (24, 28)


class SvgPushButton(QPushButton):
    """
    A standard SVG QPushButton.

    Arguments:
        width (int):
            Button width.

        height (int):
            Button height.

        svg_path (str):
            Path to the SVG file to be the default display state of the button.

        svg_path_hover (str):
            Path to the SVG file to be the hover display state of the button.
    """

    def __init__(self, parent, dimensions: BtnSize, icon: BtnIcon, clickable=True) -> None:
        super().__init__(parent)

        if not isinstance(dimensions, BtnSize):
            raise TypeError(f"Argument 'dimensions' must be of type BtnSize, not {type(dimensions)}")

        if not isinstance(icon, BtnIcon):
            raise TypeError(f"Argument 'icon' must be of type BtnIcon, not {type(icon)}")

        if not isinstance(clickable, bool):
            raise TypeError(f"Argument 'clickable' must be of type bool, not {type(clickable)}")

        self.dims = dimensions.value

        """Standard icon filename structure"""
        icon_dir = "resources/icons/"
        self.svg_path = epath(icon_dir + icon.value + ".svg")
        if clickable:
            self.svg_path_hover = epath(icon_dir + icon.value + "_hover.svg")
        else:
            self.svg_path_hover = None

        self._hover_message = None

        self.render(self.svg_path)

        self.enterEvent = lambda event: self.hover_enter(event, self.svg_path_hover)
        self.leaveEvent = lambda event: self.hover_leave(event, self.svg_path)
        if self.svg_path_hover:
            self.mousePressEvent = lambda event: self.mouse_press(event, self.svg_path)
            self.mouseReleaseEvent = lambda event: self.mouse_release(event, self.svg_path_hover)

    @property
    def hover_message(self):
        return self._hover_message

    def mouse_release(self, event, svg_path: str):
        """
        The effects of click mouse release on the button.
        """

        super().mouseReleaseEvent(event)
        self.render(svg_path)

    def mouse_press(self, event, svg_path: str):
        """
        The effects of click mouse down on the button.
        """

        super().mousePressEvent(event)
        self.render(svg_path)

    def hover_enter(self, event, svg_path: str):
        """
        The effects of entering hover for the button.
        """

        super().enterEvent(event)
        if self.svg_path_hover:
            self.setCursor(QCursor(Qt.PointingHandCursor))
        if svg_path:
            self.render(svg_path)

        if self.hover_message:
            StatusBar.message(self.hover_message)

    def hover_leave(self, event, svg_path: str) -> None:
        """
        The effects of leaving hover for the button.
        """

        super().leaveEvent(event)
        if svg_path:
            self.render(svg_path)

        # if StatusBar.status_bar(raise_on_error=False):
        #     StatusBar.message('')

    def set_hover_status_message(self, message: str):
        """
        Set the message to be displayed in the status bar when this button is hovered over.
        """

        self._hover_message = message

    def render(self, svg_path: str) -> None:
        """
        Render the SVG icon for the button.
        """

        svg_file = QFile(svg_path)
        svg_file.open(QFile.ReadOnly | QFile.Text)
        svg_renderer = QSvgRenderer(svg_file.readAll())
        svg_size = QSize(*self.dims)
        svg_pixmap = QPixmap(svg_size)
        svg_pixmap.fill(Qt.transparent)

        painter = QPainter(svg_pixmap)
        svg_renderer.render(painter)
        painter.end()

        self.setFixedSize(svg_size)
        self.setStyleSheet(
            """
            QPushButton {
                border: none;
                padding: 0px;
            }
        """
        )
        self.setIcon(QIcon(svg_pixmap))
        self.setIconSize(svg_size)



from pathlib import Path

from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices

from houdini_package_manager.meta.meta_tools import StatusBar


class Actions:
    """
    Various actions that can be performed by widgets.
    """

    def open_path(path):
        """
        Get the path that is associated with a button and open it.
        """

        if not isinstance(path, Path):
            path = Path(path)

        if not path.exists():
            StatusBar.message(f"Failed to open: {str(path)}")
            return

        QDesktopServices.openUrl(QUrl.fromLocalFile(path))
        StatusBar.message(f"Opened: {path}")



# for processing cli args
import logging
import sys

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication

from houdini_package_manager.widgets.dialogs import ErrorDialog, Updater
from houdini_package_manager.widgets.main_window import MainWindow
from houdini_package_manager.wrangle.config_control import HoudiniManager


def main(start: bool = True, headless: bool = False) -> QApplication:
    """Start the Houdini package manager app

    Create and open the Houdini package manager window

    Args:
        start (bool): Start the app. Default is True.
            If False, the app won't start but the QApplication object will still be created.
        headless (bool): The app window will not show when the app starts if True. Default is False.

    Returns:
        QApplication: PySide6.QtWidgets object
    """

    logging.debug("Houdini Package Manager - Python version:")
    logging.debug(sys.version)
    logging.debug(f"{sys.version_info}\n")

    app = QApplication(sys.argv)

    # get all the relevant houdini meta and package data for all installed versions of Houdini
    houdini_data = HoudiniManager()
    houdini_data.get_houdini_data()

    # check for no houdini install
    if not houdini_data.install_directories:
        dialog_missing_houdini = ErrorDialog(
            "No Houdini install found.\nPlease install Houdini first.\n\nHPM's functionality relies on a version of"
            " Houdini being installed to work."
        )
        dialog_missing_houdini.exec()
        return

    window = MainWindow(app, houdini_data)

    # delay updater dialog until main window shows
    QTimer.singleShot(0, show_updater)

    if not headless:
        window.show()

    if start:
        app.exec()

    return app


def show_updater():
    """
    Show the version update dialog.
    """

    updater = Updater()
    updater.check_update()


if __name__ == "__main__":
    main()



import logging
from pathlib import Path
from typing import Any, List

from PySide6.QtCore import QEvent, Qt, QTimer, QUrl
from PySide6.QtGui import QDesktopServices, QStandardItem, QStandardItemModel
from PySide6.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QComboBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSizePolicy,
    QStyledItemDelegate,
    QTableWidget,
    QWidget,
)

from houdini_package_manager.meta.meta_tools import (
    GlobalExceptionTracker,
    RateLimitError,
    RequestConnectionError,
    StatusBar,
    TableHeaders,
    TextColor,
)
from houdini_package_manager.widgets.custom_widgets import BtnIcon, BtnSize, SvgPushButton
from houdini_package_manager.wrangle.config_control import HoudiniInstall, Package
from houdini_package_manager.wrangle.url import Url


class PackageTableModel(QTableWidget):
    """
    The table widget that displays Houdini package configuration data and various
    buttons/options to navigate and manipulate them.

    Every cell of the table by default contains a QWidget container which is used
    to align and scale its inner contents. Contents can consist of text, buttons,
    dropdowns, or other functionality. If a container is not set to have anything
    in it (because nothing applies for that cell), then it will be filled with
    an empty QLabel by default, which displays as a blank cell.

    A default QWidget for every cell helps prevent errors when programmatically
    traversing multiple cells.
    """

    def __init__(self, parent, houdini_install: HoudiniInstall) -> None:
        super().__init__(parent)

        if not houdini_install.packages.pkgs:
            raise ValueError(f"No package data found for Houdini {houdini_install.version.full}")

        self.hou_version = houdini_install.version
        self.packages = houdini_install.packages.pkgs
        self.table_data = houdini_install.pkg_data_as_table_model(named=False)
        self.column_headers = houdini_install.get_labels()
        self.warnings = houdini_install.get_package_warnings()

        self.setup_table_data()
        self.fill_table_contents()

    @property
    def cell_data(self) -> List[List[Any]]:
        """
        Returns a grid of each cell's inner data for this table.

        This refers to the inner most desirable data:
        table -> cell -> QWidget container -> desired data

        This should only be called after the table has been fully setup with data
        (QWidget container and any inner data) in all cells.
        """

        data = []
        for i in range(self.rowCount()):
            data.append([])
            for j in range(self.columnCount()):
                cell = self.cellWidget(i, j).layout().itemAt(0).widget()
                data[i].append(cell)

        return data

    def setup_table_data(self) -> None:
        """
        Prepare the table data before it gets entered into the packages table.
        This method does not determine column order. This is pre-set by the table_model property in the Package class.
        """

        # insert custom index column
        self.column_headers.insert(0, "")
        for i, _row in enumerate(self.table_data):
            self.table_data[i].insert(0, str(i + 1))
        # remove default index column since it contains the corner part that can't be styled
        for i in range(self.rowCount()):
            item = self.item(i, 0)
            item.setTextAlignment(Qt.AlignCenter)
        self.verticalHeader().setVisible(False)

        self.setRowCount(len(self.table_data))
        self.setColumnCount(len(self.column_headers))

        self.setHorizontalHeaderLabels(self.column_headers)
        self.resizeColumnsToContents()  # before manual width adjustments
        self.setColumnWidth(self.column_headers.index(TableHeaders.PACKAGE.value), 180)
        self.setColumnWidth(self.column_headers.index(TableHeaders.PLUGINS.value), 200)

        # resize troublesome columns to their contents after event loop is idle (window shown)
        QTimer.singleShot(0, self.minimize_column_widths)

        style = """
            ::section {
                background-color: #3d3d3d;
                border: none;
                border-left: 1px solid #202020;
                border-right: 1px solid #202020;
            }
        """
        self.horizontalHeader().setStyleSheet(style)
        self.verticalHeader().setStyleSheet(style)

        self.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.setEditTriggers(QTableWidget.NoEditTriggers)
        self.setSelectionMode(QTableWidget.NoSelection)
        self.setShowGrid(False)

        header = self.horizontalHeader()
        header.setStretchLastSection(True)

        # expand vertical cell size
        # self.verticalHeader().setMinimumWidth(26) # default index column width
        self.verticalHeader().setDefaultSectionSize(36)  # cell height

    def minimize_column_widths(self):
        """
        Resizes certain table's column widths to their content width.

        This is to improve quality of life when using the app by working around an annoying
        feature of certain columns horizontally resizing inconsistently.

        This method really only exists to support resizing any troublesome columns that:
            - refuse to allow resizing to their contents until after the app window is shown.
            - size unpredictably/inconsistently across multiple tables, leading to data
            visually jumping around horizontally when you switch between tables.

        Add more columns to this list whenever they start to become troublesome.
        """

        column_indices = [0]  # start with custom index column
        column_indices += self._headers_to_column_index(TableHeaders.SOURCE)

        for i in column_indices:
            self.resizeColumnToContents(i)

    def setCellWidget(self, widget: QWidget) -> None:
        """
        Wrapper for the QTableWidget setCellWidget method to simplify setting cell data.
        """

        return super().setCellWidget(self.t_row, self.t_col, widget)

    def fill_table_contents(self) -> None:
        """
        Enters the packages data into the packages table.

        The QTableWidget's cells each consist of a QWidget container that aligns the
        desired widget (QLabel, QPushButton, etc.) inside it.
        (QTableWidget -> cell -> QWidget -> Any widget type)

        All cell items are center aligned by default.
        """

        columns_to_align_left = self._headers_to_column_index(
            [TableHeaders.PACKAGE, TableHeaders.AUTHOR, TableHeaders.PLUGINS]
        )
        self.t_row = 0
        self.t_col = 0

        # set table data with widgets based on the incoming data's information/data type
        for row, rowData in enumerate(self.table_data):
            self.t_row = row
            for col, value in enumerate(rowData):
                self.t_col = col

                header = self.horizontalHeaderItem(self.t_col).text()  # column title
                if header:
                    header = self._string_to_header_enum(header)  # convert header back to its enum

                # checkbox
                if header in [TableHeaders.ENABLE]:
                    widget = CellWidgets.checkbox_enable(self, value)

                # button - open a path
                elif header in [TableHeaders.SOURCE, TableHeaders.CONFIG] and value:
                    widget = CellWidgets.button_open_path(self, value, header)

                # button - sync a package's repository metadata
                elif header in [TableHeaders.SYNC] and self._current_package().remote_repo_url:
                    widget = CellWidgets.button_git_sync(self)

                # dropdown (warning button as a replacement)
                elif header in [TableHeaders.PLUGINS]:
                    widget = self._create_plugin_dropdown_widget(value)

                # index (str), string, None - strings and any remaining unset values from the table model
                elif isinstance(value, str) or not value:
                    widget = CellWidgets.label_text(value)

                # align widgets in cells
                align = None
                if self.t_col in columns_to_align_left:
                    align = Qt.AlignLeft

                self.setCellWidget(self.align_widget(widget, align))

    def _create_plugin_dropdown_widget(self, value) -> QWidget:
        """
        Creates the plugins dropdown widget.
        """

        warnings = "\n".join(list(self.warnings.values())[self.t_row])
        if warnings:
            widget = CellWidgets.button_warning(self, self.t_row, warnings, self._current_package().config_path)
        elif not value:
            widget = CellWidgets.label_no_plugin_data()
        elif len(value) > 1:
            widget = CellWidgets.combo_plugins(self, value)
        else:
            widget = CellWidgets.button_plugins(self, value[0])
            widget.clicked.connect(self.open_path)
        widget.setMinimumHeight(29)
        return widget

    def open_path(self) -> None:
        """
        Get the path that is associated with a button and open it.
        This method can handle both path Url and Path objects.
        """

        button = self.sender()
        path = button.property("path")  # Url or Path
        path_str = str(path)

        if isinstance(path, Path):
            if not path.exists():
                StatusBar.message(f"Failed to open: {path_str}")
                return
            QDesktopServices.openUrl(QUrl.fromLocalFile(path))
        elif isinstance(path, Url):
            QDesktopServices.openUrl(QUrl.fromUserInput(path_str))

        StatusBar.message(f"Opened: {path_str}")

    def enable_package(self) -> None:
        """
        This method is called when an enable checkbox is clicked.
        This sets the 'enable' variable of the relevant JSON package to the toggle value by
        triggering the packages enable attribute setter method, thus enabling or disabling
        the package.
        """

        checkbox = self.sender()
        toggle = checkbox.isChecked()
        row = self.indexAt(checkbox.parent().pos()).row()

        package = list(self.packages.values())[row]
        package.enable = toggle  # triggers setter method

        message = "Enabled" if toggle else "Disabled"
        StatusBar.message(f"{message} package: {package.pkg_name}")

    def align_widget(self, widget, align: Qt.AlignmentFlag = None) -> QWidget:
        """
        Creates a QWidget with a QHBoxLayout that positions and scales a desired widget inside.

        These are mainly sub-containers that are intended to be set inside a QTableWidget cell.
        The QWidget prevents the cell from collapsing in on the cell contents, which would hide it.

        Args:
            widget:
                The QWidget container to be placed in a QTableWidget cell.

            align:
                The type of alignment. Centered inside the cell by default.

        Returns:
            A QWidget that can be used as an argument in the self.setCellWidget method.
        """

        if not align:
            align = Qt.AlignCenter

        layout_widget = QWidget()
        layout = QHBoxLayout(layout_widget)
        layout.addWidget(widget)
        layout.setAlignment(align)
        layout.setContentsMargins(0, 0, 0, 0)
        layout_widget.setLayout(layout)
        return layout_widget

    def _current_package(self) -> Package:
        """
        Return the current Package that's data is being worked upon while the table cells are being set.
        """

        pkg = list(self.packages.values())[self.t_row]
        return pkg

    def _current_cell(self) -> tuple:
        """
        The row and column integers for the current cell that is being created while the given cell is being set.

        Returns a tuple.
        """

        return (self.t_row, self.t_col)

    def _get_cell_contents_from_same_row(self, header: TableHeaders) -> QWidget | None:
        """
        Returns the desired QTableWidget cell's widget contents (if any).

        The desired cell must be targeted:
            - from another cell within the same row.
            - during the calling cell's creation (so that the current row & column indices are correct).
            - with a given column header from TableHeaders.

        Args:
            header (TableHeaders):
                The target cell's column header Enum name.

        Returns:
            A QWidget object (the cell's inner alignment container) if one exists or None if the cell is empty.
        """

        cell_position = self._get_cell_position_from_same_row(header)
        cell_widget = self.cellWidget(*cell_position)

        # in case cell has not been set with anything
        if not cell_widget:
            message = (
                f"Table cell at position {cell_position} is missing a default inner QWidget container (cell has no"
                " data)."
            )
            logging.error(message)
            StatusBar.message(message, TextColor.ERROR)
            return None

        target_widget = cell_widget.layout().itemAt(0).widget()

        return target_widget

    def _get_cell_position_from_same_row(self, header: TableHeaders) -> tuple:
        """
        Returns the index coordinate position of the desired QTableWidget cell.

        The position is determined by iterating through the header row until
        the current header name matches the desired name.

        Args:
            header (TableHeaders):
                The target cell's column header Enum name.

        Returns:
            A tuple of the cell's coordinate position.
        """

        if not isinstance(header, TableHeaders):
            raise TypeError(f"{type(header)} must be of type TableHeaders.")

        header_row = self.horizontalHeader()
        for column in range(header_row.count()):
            if self.horizontalHeaderItem(column).text() == header.value:
                break

        cell_position = (self.t_row, column)

        return cell_position

    def _headers_to_column_index(self, headers: TableHeaders | List[TableHeaders]) -> List[int]:
        """
        Convert the column header(s) to its index.

        Returns a list of int(s).
        """

        if isinstance(headers, TableHeaders):
            headers = [headers]

        indices = []
        for head in headers:
            i = self.column_headers.index(head.value)
            indices.append(i)

        return indices

    def _string_to_header_enum(self, string_value: str) -> TableHeaders:
        """
        Converts a string to its corresponding enum member in the TableHeaders class.

        Args:
            string_value (str): The string to convert to an enum member.

        Returns:
            TableHeaders: The enum member corresponding to the string value.

        Raises:
            ValueError: If the string does not match any of the enum members' values.
        """

        for member in TableHeaders:
            if member.value == string_value:
                return member
        raise ValueError(f"{string_value} is not a valid {TableHeaders.__name__}")


class CellWidgets:
    """
    Various widget presets such as buttons, dropdowns, and labels for a packages table.
    """

    @staticmethod
    def button_warning(parent: PackageTableModel, row: int, warnings: str, config_path: Path) -> SvgPushButton:
        # Plugins: a drop down of path buttons that can be clicked.
        # A warning SVG replaces the dropdown if the package has errors that the user needs to resolve.
        # A label replaces the dropdown if there is no plugin data.

        # if the package config has problems
        button_warning = SvgPushButton(parent, BtnSize.WARNING, BtnIcon.WARNING)

        pkg_name = parent._current_package().pkg_name
        if pkg_name[-5:] != ".json":
            pkg_name += ".json"
        button_warning.set_hover_status_message(f"Can't process package, error(s) in config: {pkg_name}")
        button_warning.setToolTip(warnings)
        button_warning.setProperty("path", config_path)
        button_warning.clicked.connect(parent.open_path)

        return button_warning

    @staticmethod
    def label_no_plugin_data() -> QLabel:
        # if there's no plugin data
        label = QLabel("No plugin folder found.")
        return label

    @staticmethod
    def combo_plugins(parent, value) -> QComboBox:
        combo = DropDown(parent)
        value = [str(path) for path in value]

        model = QStandardItemModel()
        combo.setModel(model)

        for path in value:
            model.invisibleRootItem().appendRow(QStandardItem(path))

        # create a delegate that handles the paint and editorEvent events for the items in the combo box.
        # this allows the dropdown items to open their paths when they are clicked.
        delegate = QStyledItemDelegate()
        combo.setItemDelegate(delegate)

        delegate = CustomItemDelegate()
        combo.setItemDelegate(delegate)
        combo.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        combo.setStyleSheet(
            """
        QComboBox {
            text-align: left;
            background-color: #303030;
            border: none;
            padding: 5px;
        }

        QComboBox::drop-down { subcontrol-position: left; }
        QComboBox:hover { background-color: #404040; }
        """
        )

        return combo

    @staticmethod
    def button_plugins(parent, value) -> QPushButton:
        button = QPushButton(text=str(value), parent=parent)
        button.setProperty("path", value)
        button.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        button.setStyleSheet(
            """
        QPushButton {
            text-align: left;
            background-color: #303030;
            border: none;
            padding: 5px;
        }

        QPushButton:hover {
            background-color: #404040;
        }

        QPushButton:pressed { background-color: #303030; }
        """
        )

        return button

    @staticmethod
    def button_open_path(parent: PackageTableModel, path: Path | Url, button_type: TableHeaders) -> SvgPushButton:
        """
        A QPushButton that opens its file path or url when clicked.
        """

        # local path or url
        if button_type is TableHeaders.SOURCE:
            button = SvgPushButton(parent, BtnSize.CELL_TALL, BtnIcon.SOURCE_CONTROL)
        elif button_type is TableHeaders.CONFIG:
            button = SvgPushButton(parent, BtnSize.CELL_TALL, BtnIcon.FILE)

        button.setToolTip(str(path))
        button.setProperty("path", path)  # store path on button
        button.clicked.connect(parent.open_path)
        button.set_hover_status_message(f"Open: {str(path)}")
        return button

    @staticmethod
    def button_git_sync(parent: PackageTableModel) -> SvgPushButton:
        """
        A QPushButton that updates the row with the latest tag version.

        One or more requests is sent to the GitHub API to fetch the remote repo's tags.
        """

        # initialize data now during this cell's creation
        pkg = parent._current_package()
        target_cell = parent._get_cell_contents_from_same_row(TableHeaders.LATEST)
        exception_tracker = GlobalExceptionTracker()

        def update_latest_tag() -> None:
            """
            Perform all operations to update the latest tag version.

            The tag is fetched from the remote, set to the relevant cell, and finally cached in the user's json.
            """

            try:
                latest_tag = pkg._git_project.fetch_latest_remote_tag()

                if not latest_tag:
                    StatusBar.message(
                        f"Sync failed. No tag versions found on the remote repository for {pkg.pkg_name}.",
                        TextColor.ERROR,
                    )
                    return

                target_cell.setText(latest_tag)
                StatusBar.message(f"Successfully synced metadata for {pkg.pkg_name}", TextColor.SUCCESS)

            except (RateLimitError, RequestConnectionError) as e:
                exception_tracker.set_exception(e)
                StatusBar.message(str(e), TextColor.ERROR)

        button = SvgPushButton(parent, BtnSize.SQUARE_DEFAULT, BtnIcon.GIT_SYNC)
        button.clicked.connect(update_latest_tag)
        button.setToolTip("Sync metadata")
        button.set_hover_status_message(f"Sync metadata for {pkg.remote_repo_url.stem}")
        return button

    @staticmethod
    def label_text(value) -> QLabel:
        # Version, Name, Author, Date Installed: text
        item = QLabel(value)
        return item

    @staticmethod
    def checkbox_enable(parent: PackageTableModel, value) -> QCheckBox:
        # Enable: togglable checkbox
        checkbox = QCheckBox()
        checkbox.clicked.connect(parent.enable_package)
        checkbox.setCheckState(Qt.Checked if value else Qt.Unchecked)
        return checkbox


class DropDown(QComboBox):
    """
    A dropdown list of items that ignores scroll-wheel scrolling.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setStyleSheet("QComboBox:focus { border: 2px solid #100000; }")

    def wheelEvent(self, event):
        event.ignore()


class CustomItemDelegate(QStyledItemDelegate):
    """
    CustomItemDelegate is a custom QStyledItemDelegate that handles the paint and
    editorEvent events for the items in a QComboBox. It overrides the paint and editorEvent
    methods of the QStyledItemDelegate to customize the appearance and behavior of the items.
    """

    def paint(self, painter, option, index):
        option.text = index.data()
        super().paint(painter, option, index)

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.Type.MouseButtonPress and event.button() in [
            Qt.LeftButton,
            Qt.RightButton,
            Qt.MiddleButton,
        ]:
            QDesktopServices.openUrl(QUrl.fromLocalFile(index.data()))
            return True
        return super().editorEvent(event, model, option, index)



import logging
import shutil
from pathlib import Path
from typing import Dict, List

from PySide6.QtCore import Qt, QUrl
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QStackedWidget,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from houdini_package_manager.meta.meta_tools import (
    GlobalExceptionTracker,
    RateLimitError,
    RequestConnectionError,
    StatusBar,
    TableHeaders,
    TextColor,
)
from houdini_package_manager.widgets.custom_widgets import BtnIcon, BtnSize, SvgPushButton
from houdini_package_manager.widgets.packages_table import PackageTableModel
from houdini_package_manager.wrangle.config_control import HoudiniManager, Package


class PackagesWidget(QWidget):
    """
    The packages layout that displays the main packages table and relevant dropdowns and buttons.

    Arguments:
        parent:
            The parent widget.

        table_data (HoudiniManager):
            The data set to populate the QWidgetTable. Data can consist of multiple sets of packages
            for different installed versions of Houdini.

        versions (List[str]):
            The list of ordered version numbers that will determine which set of package data is shown in the table.

        tabs (QTabWidget):
            The parent QTabWidget. Used for switching tabs based on widget interaction in this QWidget.
    """

    def __init__(self, parent, table_data: HoudiniManager, hou_versions: List[str], tabs: QTabWidget) -> None:
        super().__init__(parent)

        self.parent_tabs = tabs
        self.table_data = table_data
        self.hou_versions = hou_versions

        self.version_labels = ["Houdini " + version for version in self.hou_versions]
        self._current_table_version = self.hou_versions[0]

        # LABEL - HOUDINI VERSION DROPDOWN
        label_version_dropdown = QLabel("HOUDINI VERSIONS")
        label_version_dropdown.setStyleSheet("QLabel { color: #ff6600 }")

        # BUTTON - ADD PACKAGE
        button_add_package = SvgPushButton(self, BtnSize.WIDE_LARGE, BtnIcon.ADD_PKG)
        button_add_package.clicked.connect(self.add_packages)

        # DROPDOWN - HOUDINI VERSION
        self.combo_version = QComboBox()
        self.combo_version.addItems(self.version_labels)
        self.combo_version.activated.connect(self.switch_package_table)
        self.combo_version.setMinimumHeight(29)
        self.combo_version.setMinimumWidth(155)

        self.combo_version.setStyleSheet(
            """
        QComboBox {
            text-align: left;
            background-color: #303030;
            border: none;
            padding: 5px;
        }

        QComboBox:hover { background-color: #404040; }
        QComboBox::drop-down { subcontrol-position: left; }
        """
        )

        # BUTTON - HOUDINI VERSION FOLDER
        self.button_version = SvgPushButton(self, BtnSize.FOLDER, BtnIcon.FOLDER)
        path = self.current_packages_directory()
        self.button_version.set_hover_status_message(f"Open: {path}")
        self.button_version.setToolTip("Open packages folder")
        self.button_version.setProperty("path", path)
        self.button_version.clicked.connect(self.open_path)

        # BUTTONS - PACKAGE OPTIONS
        # copy all the packages in the current table to another houdini version
        self.button_git_sync = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.GIT_SYNC)
        self.button_git_sync.set_hover_status_message(
            "Sync all package metadata from remote repositories for these plugins."
        )
        self.button_git_sync.setToolTip("Sync these plugins' metadata")
        self.button_git_sync.clicked.connect(self.fetch_all_package_remote_metadata)

        self.button_copy = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.MIGRATE)
        self.button_copy.set_hover_status_message("Copy all packages to other Houdini versions.")
        self.button_copy.setToolTip("Copy all packages to other Houdini versions")
        self.button_copy.clicked.connect(self.migrate_packages)

        self.button_refresh = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.REFRESH)
        self.button_refresh.set_hover_status_message(f"Refresh Houdini {self.current_table_version} packages.")
        self.button_refresh.setToolTip("Refresh packages for current table")
        self.button_refresh.clicked.connect(self.refresh_table)

        self.button_refresh_all = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.REFRESH_ALL)
        self.button_refresh_all.set_hover_status_message("Refresh all packages for all installed versions of Houdini.")
        self.button_refresh_all.setToolTip("Refresh all packages for all installed versions of Houdini")
        self.button_refresh_all.clicked.connect(self.refresh_all_tables)

        # TABLE - PACKAGE DATA
        current_hou_version = self.table_data.hou_installs[self.hou_versions[0]]
        if current_hou_version.packages.pkgs:
            table = PackageTableModel(self, current_hou_version)
            table.setStyleSheet("QTableWidget {border: none;}")
        else:
            table = self._no_packages_found()

        self.stacked_widget = QStackedWidget()
        self.stacked_widget.addWidget(table)
        self.stacked_widget.setStyleSheet("QStackedWidget {border: 1px solid grey;}")

        # keep track of loaded package tables in the order they are added to stacked_widget
        self.loaded_stacked_widgets_in_order_loaded = [self.hou_versions[0]]

        # CREATE LAYOUTS
        self.layout_main = QVBoxLayout()
        layout_secondary_header = QHBoxLayout()
        layout_table_options = QHBoxLayout()
        layout_version_buttons = QHBoxLayout()
        layout_package_buttons = QHBoxLayout()

        # SET LAYOUTS
        self.layout_main.addLayout(layout_secondary_header)
        self.layout_main.addLayout(layout_table_options)
        self.layout_main.addWidget(self.stacked_widget)

        layout_secondary_header.addWidget(label_version_dropdown)
        layout_secondary_header.addWidget(button_add_package)
        layout_secondary_header.setAlignment(label_version_dropdown, Qt.AlignBottom)

        layout_table_options.addLayout(layout_version_buttons)
        layout_version_buttons.addWidget(self.combo_version)
        layout_version_buttons.addWidget(self.button_version)
        layout_table_options.addLayout(layout_package_buttons)
        layout_package_buttons.addWidget(self.button_git_sync)
        layout_package_buttons.addWidget(self.button_copy)
        layout_package_buttons.addWidget(self.button_refresh)
        layout_package_buttons.addWidget(self.button_refresh_all)
        layout_version_buttons.setAlignment(Qt.AlignLeft)
        layout_table_options.setAlignment(layout_package_buttons, Qt.AlignRight)

    @property
    def current_table_version(self) -> str:
        """
        Returns the houdini version number of the current displayed package table.
        """

        self._current_table_version = self.combo_version.currentText().split(" ")[-1]
        return self._current_table_version

    @property
    def current_table(self) -> PackageTableModel:
        """
        Returns the currently displayed PackageTableModel package table.
        """

        table = self.stacked_widget.currentWidget()
        return table

    @property
    def loaded_tables(self) -> List[PackageTableModel]:
        """
        Returns a list of currently loaded PackageTableModel package tables.

        Tables that aren't loaded are ones that the user has not navigated to yet,
        which hasn't triggered their load.
        """

        tables = []
        for i in range(self.stacked_widget.count()):
            tables.append(self.stacked_widget.widget(i))
        return tables

    def add_packages(self) -> None:
        """
        Switch to the tab for creating packages from local plugins.
        Choosing the tab index to switch to is determined by a tab's
        matching "id" property.
        """

        target_id = "local_plugin_adder"
        tab_widget_id = []
        for i in range(self.parent_tabs.count()):
            tab_widget_id.append(self.parent_tabs.widget(i).property("id"))

        target_index = tab_widget_id.index(target_id)

        self.parent_tabs.setCurrentIndex(target_index)

    def open_path(self) -> None:
        """
        Get the path that is associated with a button and open it.
        """

        button = self.sender()
        path = button.property("path")

        if not isinstance(path, Path):
            path = Path(path)

        if not path.exists():
            StatusBar.message(f"Failed to open: {str(path)}")
            return

        QDesktopServices.openUrl(QUrl.fromLocalFile(path))
        StatusBar.message(f"Opened: {path}")

    def switch_package_table(self) -> None:
        """
        Load a set of package data into a table, add it to the QStackedWidget, and switch to it.
        If the set of data was already loaded then switch to it instead of loading it again.
        """

        # MANUAL UPDATING OF PACKAGE TABLE INDEPENDENT WIDGETS
        # update refresh button hover status bar message
        self.button_refresh.set_hover_status_message(f"Refresh Houdini {self.current_table_version} packages.")
        # update houdini packages folder button location
        self.button_version.setProperty("path", self.current_packages_directory())
        self.button_version.set_hover_status_message(f"Open: {self.current_packages_directory()}")

        # if the table widget has already been added, switch to it
        if self.current_table_version in self.loaded_stacked_widgets_in_order_loaded:
            index = self.loaded_stacked_widgets_in_order_loaded.index(self.current_table_version)
            self.stacked_widget.setCurrentIndex(index)
            return

        configs = self.table_data.hou_installs[self.current_table_version].packages.pkgs
        if configs:
            widget_contents = PackageTableModel(self, self.table_data.hou_installs[self.current_table_version])
            widget_contents.setStyleSheet("QTableWidget {border: none;}")
        else:
            widget_contents = QLabel("No packages found")
            widget_contents.setAlignment(Qt.AlignCenter)

        self.stacked_widget.addWidget(widget_contents)
        self.stacked_widget.setCurrentWidget(widget_contents)
        self.loaded_stacked_widgets_in_order_loaded.append(self.current_table_version)

    def refresh_table(self, version: str = None, status: bool = True) -> None:
        """
        Refresh all the package data for the desired table.
        If version is specified, the currently displayed table's data will be refreshed.

        Arguments:
            version (str):
                The version number from the QComboBox list of the package table to refresh.
                If version is None, the currently displayed package data and table in the QStackedWidget will be refreshed.
                If version is a valid version, that loaded package data and table will be refreshed.

            status (bool):
                Whether the status bar should print a successful package refresh message.
                Suppressing this is useful when multiple package tables need to be refreshed all at once.
                Default is True.
        """

        refresh_version = version if version else self.current_table_version

        # refresh config data for current package set
        self.table_data.get_houdini_data(refresh_version)

        configs = self.table_data.hou_installs[refresh_version].packages.pkgs
        if configs:
            widget_contents = PackageTableModel(self, self.table_data.hou_installs[refresh_version])
            widget_contents.setStyleSheet("QTableWidget {border: none;}")
        else:
            if status:
                StatusBar.message(f"No packages found for Houdini {refresh_version}.")
            widget_contents = self._no_packages_found()

        current_index = self.stacked_widget.currentIndex()

        # refresh QStackWidget widget (if it has been created)
        if refresh_version not in self.loaded_stacked_widgets_in_order_loaded:
            return

        target_index = self.loaded_stacked_widgets_in_order_loaded.index(
            refresh_version
        )  # index position in self.stacked_widget, NOT the QComboBox
        target_widget = self.stacked_widget.widget(target_index)

        self.stacked_widget.insertWidget(target_index, widget_contents)
        self.stacked_widget.removeWidget(target_widget)
        self.stacked_widget.setCurrentIndex(current_index)

        if status:
            StatusBar.message(f"Refreshed packages for Houdini {refresh_version}.")

    def _no_packages_found(self):
        """
        Returns a QLabel that's meant to replace a table when there is no package data to fill it.
        """

        label = QLabel("No packages found")
        label.setAlignment(Qt.AlignCenter)
        return label

    def refresh_all_tables(self) -> None:
        """
        Refresh the package data for all versions of Houdini.
        First refresh the data for every loaded table, then reload the rest
        of package config data that hasn't yet been converted to a table. This
        is to prevent having to load tables for the entire list of Houdini installs.
        """

        # # lazy method - refresh everything (less efficient)
        # # refresh all config data
        # self.table_data.get_houdini_data()
        # # refresh all loaded widgets in QStackedWidget
        # # DOESNT WORK CURRENTLY ANYWAYS SINCE CAN'T REFRESH A NON-LOADED WIDGET
        # for version in self.versions:
        #     self.refresh_table(version)

        # refresh loaded tables
        for version in self.loaded_stacked_widgets_in_order_loaded:
            self.refresh_table(version)

        # reload package config data for packages not loaded into tables.
        not_loaded_versions = list(set(self.hou_versions).difference(set(self.loaded_stacked_widgets_in_order_loaded)))
        for version in not_loaded_versions:
            self.table_data.get_houdini_data(version)

        StatusBar.message("Refreshed all package data and tables.")

    def fetch_all_package_remote_metadata(self) -> None:
        """
        Fetch the remote metadata for every package of this version of Houdini.

        This is done by programmatically "clicking" every git sync button in the
        table across all the rows.

        Exceptions that occur during metadata syncing across multiple rows are
        tracked by a global exception tracker and then handled in this method
        because programmatically triggering buttons' signals prevents
        exceptions from traversing up the call stack to this method where
        they need to be caught. The global exception tracker is the
        workaround for this.
        """

        logging.debug(f"Started full table metadata GitHub sync for Houdini {self.current_table_version}")

        table = self.current_table
        cell_data = table.cell_data
        git_fetch_col_index = table._headers_to_column_index(TableHeaders.SYNC)[0]
        tracker = GlobalExceptionTracker()

        for row in cell_data:
            cell_item = row[git_fetch_col_index]
            if isinstance(cell_item, SvgPushButton):  # checking button exists in cell
                cell_item.clicked.emit()  # trigger signal
                exception = tracker.get_exception()  # recheck if hidden exception occured
                if exception:
                    break

        if not exception:
            logging.debug(
                f"Completed full table metadata GitHub sync for Houdini {self.current_table_version} with no errors."
            )
            StatusBar.message(
                f"Synced all metadata for applicable Houdini {self._current_table_version} packages.", TextColor.SUCCESS
            )
        elif exception:
            logging.error(
                f"Failed to complete full table metadata GitHub sync for Houdini {self.current_table_version}"
            )

            if isinstance(exception, RateLimitError):
                StatusBar.message(
                    "GitHub API rate limiting occured at some point during table metadata sync. Some packages weren't"
                    " synced. Try again later.",
                    TextColor.ERROR,
                )
            elif isinstance(exception, RequestConnectionError):
                StatusBar.message(
                    "A connection error occured trying to reach the GitHub API during table metadata sync. Try again.",
                    TextColor.ERROR,
                )

        # TODO: print total number of pkgs whose tag lists have been updated

    def migrate_packages(self) -> None:
        """
        Make copies of all the packages in the currently displayed table and put them
        in the target packages directory of the target installed versions of Houdini.
        """

        if len(self.hou_versions) <= 1:
            StatusBar.message("No other Houdini versions to copy packages to.", TextColor.ERROR)
            return

        checkbox_version_options = list(self.hou_versions)
        checkbox_version_options.remove(self.current_table_version)
        checkbox_version_options = [f"Houdini {version}" for version in checkbox_version_options]

        # identify any potential file overwrite conflicts for different houdini versions
        current_package_paths = [package.config_path for package in self.current_packages()]
        if not current_package_paths:
            StatusBar.message(f"No packages in Houdini {self.current_table_version} to copy.")
            return

        file_conflicts = self.find_file_conflicts()
        dialog = CheckboxDialog(checkbox_version_options, len(current_package_paths), file_conflicts)
        dialog.setStyleSheet(
            """
            * {
                background-color: #303030;
                color: white;
            }
        """
        )
        result = dialog.exec()
        if result == 0:
            return

        target_versions = dialog.enabled_checkboxes()
        target_versions = [version.split(" ")[-1] for version in target_versions]
        target_packages_dirs = [
            self.table_data.hou_installs[version].packages.packages_directory for version in target_versions
        ]

        # copy files
        for dest in target_packages_dirs:
            for file_path in current_package_paths:
                shutil.copy(file_path, dest)

        StatusBar.message(
            f"Copied {len(current_package_paths)} packages from {self.combo_version.currentText()} to"
            f" {len(target_packages_dirs)} other Houdini installs."
        )

        # refresh packages table for target houdini versions
        for version in target_versions:
            self.refresh_table(version, status=False)

    def find_file_conflicts(self) -> Dict[str, List[Path]]:
        """
        Search for file conflicts in other Houdini packages directories by comparing
        all the Package configs in the current houdini version with all the other houdini version Package configs.
        This is useful for if config files from one houdini version need to be moved to another.

        Returns a dict of all other houdini versions where each value is a list of package Paths.
        """

        # check the package tables that aren't the one that's currently loaded
        other_versions = list(self.hou_versions)
        other_versions.remove(self.current_table_version)

        # identify any potential file overwrite conflicts for different houdini versions
        current_package_paths = [package.config_path for package in self.current_packages()]
        other_package_paths = {}
        for version, pkgs in self.get_packages(other_versions).items():  # sort package paths to each houdini version
            pkg_paths = []
            for pkg in pkgs.values():
                pkg_paths.append(pkg.config_path)
            pkg_paths = [path.name for path in pkg_paths]
            other_package_paths[version] = pkg_paths

        file_conflicts = {}
        for version, paths in other_package_paths.items():
            file_conflicts[version] = []
            for pkg_path in current_package_paths:
                if pkg_path.name in paths:
                    file_conflicts[version].append(pkg_path)

        return file_conflicts

    def current_packages(self) -> List[Package]:
        """
        Return a list of the Package objects assosciated with the currently loaded table.
        """

        current_table = self.current_table
        if not hasattr(current_table, "packages"):
            return []
        packages = list(current_table.packages.values())
        return packages

    def get_packages(self, versions: str | List[str] = None) -> Dict[str, Package]:
        """
        Return a dict of the package data for the all the tables.
        The key is the Houdini version number.
        The value is the Package object.

        Arguments:
            versions (str | List[str]):
                The Houdini versions to get packages from.
                If no version is provided, the packages for all Houdini versions will be returned.
        """

        all_versions = self.table_data.hou_installs
        if not versions:
            return all_versions

        if isinstance(versions, str):
            versions = [versions]

        #  isolate target versions
        target_versions = {}
        for key, value in all_versions.items():
            if key in versions:
                target_versions[key] = value

        # get the package data for each houdini install version
        package_data = {}
        for key, houInstall in target_versions.items():
            package_data[key] = houInstall.packages.pkgs

        return package_data

    def current_packages_directory(self) -> Path:
        """
        Return the packages directory of the currently loaded houdini version.
        """

        current_version = self.current_table_version
        path = self.table_data.hou_installs[current_version].packages.packages_directory
        return path


class CheckboxDialog(QDialog):
    """
    The dialog box for copying packages from one Houdini version to another.
    """

    def __init__(self, checkbox_data, package_count, file_conflicts, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Copy packages")

        self.file_conflicts = file_conflicts  # all possible conflicts
        self.checked_file_conflicts = 0  # accumulated number of conflicts from only checked houdini versions
        self.checkboxes = []

        label_pkg_num = QLabel(f"Copy {package_count} packages to:")
        self.label_file_overwrites = QLabel(f"{self.checked_file_conflicts} files will be overwritten.")

        layout_versions = QVBoxLayout()
        self.checkboxes = []
        for option in checkbox_data:
            checkbox = QCheckBox(option)
            checkbox.clicked.connect(self.update_file_conflicts)
            hbox = QHBoxLayout()
            hbox.addWidget(checkbox)
            hbox.setContentsMargins(0, 0, 0, 0)
            hbox.setAlignment(Qt.AlignLeft)
            layout_versions.addLayout(hbox)
            self.checkboxes.append(checkbox)
        layout_versions.setContentsMargins(0, 10, 0, 10)

        # button box
        button_box = QDialogButtonBox(Qt.Horizontal)
        button_box.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        layout_main = QVBoxLayout()
        layout_main.addWidget(label_pkg_num)
        layout_main.addLayout(layout_versions)
        layout_main.addWidget(self.label_file_overwrites)
        layout_main.addWidget(button_box)
        layout_main.setAlignment(button_box, Qt.AlignCenter)
        self.setLayout(layout_main)

    def update_file_conflicts(self) -> None:
        """
        Updates the displayed total number of file conflicts in the dialog box's warning.

        This is done by checking for conflicts for each selected version of houdini.
        """

        self.checked_file_conflicts = 0
        data = self.get_checkbox_data()

        i = 0
        for paths in self.file_conflicts.values():
            if data[i]:
                self.checked_file_conflicts += len(paths)
            i += 1

        self.label_file_overwrites.setText(f"{self.checked_file_conflicts} files will be overwritten.")

    def get_checkbox_data(self) -> List[tuple]:
        """
        Returns a list of tuples containing this dialog's checkbox labels and whether they are checked.
        """

        return [checkbox.isChecked() for checkbox in self.checkboxes]

    def enabled_checkboxes(self) -> List[str]:
        """
        Returns a list of strings of the enabled checkbox labels' text.
        """

        selected_checkboxes = []
        for checkbox in self.checkboxes:
            if checkbox.isChecked():
                selected_checkboxes.append(checkbox.text())

        return selected_checkboxes



import sys
from pathlib import Path


def _base_path() -> Path:
    """
    Return the absolute base Path of the current executing environment.
    This allows resource files to be located when the project is running from either the
    development environment or pyinstaller build environment.
    """

    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        base_path = Path(getattr(sys, "_MEIPASS", Path(__file__).resolve().parent))
        return base_path

    base_path = Path(__file__).resolve().parent
    return base_path


def epath(path: str, string=False) -> Path:
    """
    Combine the base path and a given relative path to produce a complete environment path.
    Returns a Path by default. Returns a string if string arg is True.
    """

    path = Path(_base_path(), path)

    if string:
        return str(path)

    return path



from pathlib import Path


class PackageTemplates:
    """
    Various package template formats.
    """

    @staticmethod
    def standard(path_to_plugin: Path):
        """
        A standard package template.

        Houdini will recognize:
            HOUDINI_PATH
            hpath (Shortcut to set HOUDINI_PATH. "path" has been deprecated in favor of hpath.)
            HOUDINI_TOOLBAR_PATH
            HOUDINI_OTLSCAN_PATH
            PYTHONPATH
            enable
            recommends

        Note: Do not include empty values for some keys, like recommends,
        because it was cause older versions of Houdini to crash on startup.
        """

        if not path_to_plugin.exists():
            raise FileNotFoundError(f"Does not exist: {path_to_plugin}")

        """
        The following package template is designed to be as simple and
        functional as possible. It leverages the behavior of hconfig.exe,
        which does not overwrite the HOUDINI_PATH environment variable
        whenever a new instance of HOUDINI_PATH is found in another JSON
        config. Instead, it prepends its value to previously set HOUDINI_PATH
        variables in other package JSON config files by default. Other
        environment variable key names will have their values overwritten
        if they are found in multiple package JSON config files.

        Extra paths that point to a plugin's additional folders, such as
        toolbars, OTLs, and scripts, are not required in the following
        template since Houdini automatically searches for them. As
        long as the path to the main plugin directory is included in
        the config and the plugin directory follows the standard package
        structure, the HOUDINI_PATH environment variable is sufficient
        to ensure that a typical plugin and all its components are
        loaded correctly by Houdini.

        An additional benefit of including the plugin path as a value
        of HOUDINI_PATH is that HOUDINI_PATH can be used as a variable
        ("$HOUDINI_PATH") in other parts of the JSON config. hconfig
        will recognize this when parsing the config.
        """
        package_config = {"env": [{"HOUDINI_PATH": str(path_to_plugin)}]}

        return package_config



import contextlib
import logging
import re
from pathlib import Path
from typing import List

import git
import git.exc
import requests

from houdini_package_manager.meta.meta_tools import RateLimitError, RequestConnectionError, UserDataManager
from houdini_package_manager.wrangle.url import Url


class GitProject:
    """
    A git-based project that consists of a local git repository and remote git repository for a single plugin.

    Arguments:
        get_remote (bool):
            Whether or not to fetch the local repo's remote repo data immediately after initializing the local repo data.
            The default is False. This allows the data to be fetched later at a more convenient time via user request.
    """

    def __init__(self, path: Path = None, get_remote=False) -> None:
        if not isinstance(path, Path) and path is not None:
            raise TypeError("Project path must be a Path or None.")

        if isinstance(path, Path) and not path.exists():
            raise FileNotFoundError(f"{path} does not exist.")

        self.local_path = None
        self.remote_url = None
        self.init_path(path)

        self.name = None
        self.owner = None

        self.user_data_manager = UserDataManager()

        self._local_git_repo = None  # complex git module object
        self.init_local_git_repo()
        self.local = Local()  # contains self.tags
        self.remote = Remote()  # contains self.tags
        self.init_repo_data(get_remote)

    def init_repo_data(self, get_remote=False) -> None:
        """
        Extracts any useful metadata from the local git repo and cached remote user data.

        If get_remote is True then the remote repo's data will be fetch and stored locally.
        The default is False.
        """

        """
        TODO:
        - There seems to be some duplicate tag setting being done (might be ok if its local & remote separately),
        due to logs reporting: 'Updated list of known remote tags to: ...' twice.
        If not a problem, document it and debug loggers to specify local vs remote data setting.

        COMPARE LOCAL FETCHED TAGS TO THE CACHED REMORE TAGS TO SEE WHAT'S THE TRUE LATEST TAG. This will account for a case
        where a repo gets updated locally to a new tag but the remote hasn't yet been updated.
        """

        if self._local_git_repo:
            self.extract_local_data()

            self.local.tags = self.extract_latest_tag_from_local_repo()
            if get_remote:
                self.remote.tags = self.fetch_remote_tags()

        cache = self.user_data_manager.get_entry(self.name)
        if cache:
            self.remote.tags = cache["tags"]

    def init_path(self, path) -> None:
        """
        Determines what type the incoming path is and sets the local and remote repo path attributes accordingly.
        """

        if isinstance(path, Path):
            self.local_path = path
            return

        if isinstance(path, Url):
            self.remote_url = path

    def extract_local_data(self) -> None:
        """
        Extract useful data from the local git module repo.
        """

        self._find_remote_repo_url()
        self._extract_owner()

    def extract_latest_tag_from_local_repo(self) -> str:
        """
        Extract the latest tag version from the local repo.
        Returns a string if successful.
        Returns None if not successful.
        """

        if self.local_path:
            try:
                return self._local_git_repo.git.describe(tags=True, abbrev=0)
            except git.exc.GitCommandError:
                pass
        return None

    def init_local_git_repo(self) -> None:
        """
        Create the local third party Repo object from the local path using the git module.
        If there is no .git folder present within the path's location, then no Repo will be created.

        The Repo object contains more data than is needed for HPM so we extract only what we need and store
        it in simplified custom repo objects.
        """

        if self.local_path:
            with contextlib.suppress(git.exc.InvalidGitRepositoryError):
                self._local_git_repo = git.Repo(self.local_path)

    def _extract_owner(self):
        """
        Extract the user name and repo name from the remote url.
        """

        if self.remote_url:
            split_url = str(self.remote_url).split("/")
            self.owner = split_url[-2]
            self.name = split_url[-1].split(".")[0]

    def _find_remote_repo_url(self) -> Url | None:
        """
        Find the plugin's remote source control repository URL on the local system.
        This is done by extracting the URL from the plugin path's .git folder, if it exists.

        Returns a Url object of the remote repo if it is found.
        Returns None if no Url is found.
        """

        if self.remote_url or not self.local_path:
            return

        try:
            self._local_git_repo = git.Repo(self.local_path)
        except git.exc.InvalidGitRepositoryError:
            return

        remote_urls = [remote.url for remote in self._local_git_repo.remotes]
        if remote_urls:
            self.remote_url = Url(remote_urls[0])

    def fetch_latest_remote_tag(self) -> str | None:
        """
        Gets the latest tag version from the remote repo.

        First fetches all the remote tags from a GitHub API request, then grabs the first tag.

        Returns a string if successful.
        Returns None if unsuccesful.
        """

        if self.remote_url is None:
            logging.warning("Repo URL is None (Package most likely has no remote repo). Ignoring API request command.")
            return None

        self.fetch_remote_tags()
        return self.remote.tag_latest

    def fetch_remote_tags(self, paginate=True, cache=True) -> List[str] | list:
        """
        Fetches tags from the remote repo and caches them.

        GitHub's API for fetching tags prevents requesting >100 tags in one request.
        Multiple requests and pagination is required to fetch all of a remote repo's
        tags if there are >100.

        Caching means:
            1. Storing the fetched tags locally in the user's json data.
            2. Updating the repo object (self.remote.tags) that mimics the remote tag state.

        Arguments:
            paginate (bool):
                Whether or not to keep making API requests as many times as necessary
                to fetch every possible tag via pagination. This will be necessary if more
                than the max number of tags per page (100) is required to be fetched.

            cache (bool):
                The tags that get fetched from the remote repo will be cached.
                The default is True.

        Returns:
            A list of strings if successful.
            An empty list if unsuccessful.

        Raises:
            RateLimitError: Fetch failed! Rate limited while fetching tags from '{api_tags_url}'. Status code: {response_tags.status_code}.
        """

        if self.remote_url is None:
            logging.warning("Repo URL is None (Package most likely has no remote repo). Ignoring API request command.")
            return []

        api_tags_url = f"https://api.github.com/repos/{self.owner}/{self.name}/tags"

        tags = self._fetch_all_pages(api_tags_url) if paginate else self._fetch_single_page(api_tags_url)

        if cache:
            logging.debug(f"{self.name} current list of known remote tags: {self.remote.tags}")
            self.remote.tags = tags
            self.user_data_manager.update_tags(self.name, self.remote.tags)

        return tags

    def _fetch_single_page(self, api_tags_url: str) -> List[str]:
        """
        Makes one single API request for tags (one page).

        Returns:
            A maximum of 100 tags, even if there are more on the remote repo.
            If rate limiting occurs, a RateLimitError exception occurs and no tags are returned.
        """

        tags_per_request = 100  # GitHub's API per_page max is 100
        response_tags = self._make_request(api_tags_url, params={"per_page": tags_per_request})

        if response_tags:
            tags = response_tags.json()
            tags = [tag["name"] for tag in tags]  # convert response to simple list of versions
            return tags

        return []

    def _fetch_all_pages(self, api_tags_url: str) -> List[str]:
        """
        Continually makes API requests for tags (multiple pages) until all tags are fetched.

        Returns:
            All possible tags from the remote repo.
            If rate limiting occurs, a RateLimitError exception occurs and no tags are returned.
        """

        tags_per_request = 100  # GitHub's API per_page max is 100
        pagination_page = 1  # starting page number for pagination
        all_tags = []

        while True:
            response_tags = self._make_request(
                api_tags_url, params={"per_page": tags_per_request, "page": pagination_page}
            )

            if response_tags:
                tags = response_tags.json()
                if not tags:  # no more tags to fetch
                    break
                all_tags.extend(tag["name"] for tag in tags)

                if len(tags) < tags_per_request:  # no more tags on future page requests
                    break

                pagination_page += 1
            else:
                break

        return all_tags

    def _make_request(self, url: str, params: dict) -> requests.Response | None:
        """
        Makes a GET request to the specified URL with given parameters and handles the response.

        This method handles the common logic for making a GET request to the GitHub API, checking
        for rate limiting, and logging appropriate messages based on the response status code.

        Arguments:
            url (str): The URL to send the GET request to.
            params (dict): The query parameters to include in the GET request.

        Returns:
            requests.Response: The response object if the request is successful.
            None: If the request fails with a status code other than 200 or 403.

        Raises:
            RateLimitError: If the request fails due to rate limiting (status code 403).
        """

        request_timeout = 5  # seconds

        try:
            response = requests.get(url, params=params, timeout=request_timeout)
        except requests.exceptions.ConnectionError as e:
            message = f"Fetch failed! Unable to establish connection to {url}"
            logging.error(message)
            logging.error(e)
            raise RequestConnectionError(message) from e

        if response.status_code == 200:  # success
            return response
        elif response.status_code == 403:  # rate limited
            message = (
                f"Fetch failed! Rate limited while fetching tags from '{url}'. Status code: {response.status_code}."
            )
            logging.warning(message)
            raise RateLimitError(message)
        else:
            logging.warning(
                f"Failed to fetch tags for {self.owner}/{self.name} at '{url}'. Status code:"
                f" {response.status_code}. (Package most likely has no remote repo.)"
            )
            return None


class _Repo:
    def __init__(self, tags: list | str | None) -> None:
        """
        The template for local and remote repository classes.

        This functions as a simplified repository object compared to the git module Repo object type.

        Arguments:
            tags:
                A string or list of strings of tag names.
        """

        self._tags = []
        self.tags = tags  # trigger setter

    @property
    def tag_latest(self) -> str:
        """
        The latest tag version.
        """

        if self.tags:
            return self.tags[0]
        return None

    @property
    def tags(self) -> List[str] | List[None]:
        """
        The list of known tags.
        """

        return self._tags

    @tags.setter
    def tags(self, new_tags: str | List[str] | None) -> None:
        """
        Sets a list of the known (given) tag version(s).

        Gracefully merges new tags with old tags.
        If continuity between new and old tags can't be found, the old tags are replaced with the new tags.
        """

        if new_tags is None:
            return
        elif isinstance(new_tags, list) and not new_tags:  # empty list
            return
        elif not isinstance(new_tags, str) and not all(isinstance(item, str) for item in new_tags):
            raise TypeError(f"tags {type(new_tags)} must be a str or list of str.")

        if isinstance(new_tags, str):
            new_tags = [new_tags]

        if self._tags == new_tags:
            logging.debug("Skipping tag list merge. Fetched tags list is identical to known list.")
            return

        if not self._tags:
            self._tags = new_tags
        else:
            self._tags = self._merge_version_lists(new_tags)
        logging.debug(f"Updated list of known remote tags to: {self._tags}")

    def _version_key(self, version):
        """
        Extracts numeric and non-numeric parts separately.
        """

        return [int(part) if part.isdigit() else part for part in re.split("([0-9]+)", version)]

    def _merge_version_lists(self, new_tags):
        """
        Gracefully merges the lists of new tags with old tags if possible.

        This is done by looking for common tag versions between the current and the fetched lists.
        If there are common tags, the two lists are joined with no duplicates and returned.
        If continuity between new and old tags can't be found, the old tags are replaced with the new tags.
        """

        # use sets for duplicate checking
        old_versions_set = set(self.tags)
        new_versions_set = set(new_tags)

        common_versions = old_versions_set & new_versions_set

        if common_versions:
            # merge and sort unique versions if there are common versions
            combined_versions = sorted(old_versions_set | new_versions_set, key=self._version_key, reverse=True)
            logging.debug("Common tag versions found between known and fetched tag lists. Merged lists.")
            return list(combined_versions)
        else:
            return new_tags


class Local(_Repo):
    def __init__(self, tags=None) -> None:
        super().__init__(tags)


class Remote(_Repo):
    def __init__(self, tags=None) -> None:
        super().__init__(tags)



import logging

from PySide6.QtCore import Qt, QUrl
from PySide6.QtGui import QDesktopServices, QIcon
from PySide6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QStatusBar,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from houdini_package_manager import __version__
from houdini_package_manager.meta.meta_tools import StatusBar
from houdini_package_manager.utils import epath
from houdini_package_manager.widgets.add_packages_layout import LocalPackageAdderWidget
from houdini_package_manager.widgets.custom_widgets import BtnIcon, BtnSize, SvgPushButton
from houdini_package_manager.widgets.packages_layout import PackagesWidget
from houdini_package_manager.wrangle.config_control import HoudiniManager


class MainWindow(QMainWindow):
    """
    The main window containing all of HPM's functionality.
    """

    def __init__(self, app: QApplication, houdini_data: HoudiniManager) -> None:
        super().__init__()

        self.app = app  # declare an app member
        self.houdini_data = houdini_data

        TITLE = f"Houdini Package Manager {__version__}"
        self.setWindowTitle(TITLE)

        self.setWindowIcon(QIcon(epath("resources/icons/hpm_icon.svg", True)))

        self.resize(1250, 600)

        self.setStyleSheet(
            """
                background-color: #202020;
                color: white;
                font-size: 16px;
                font-family: Lato;
                font-weight: 100;
            """
        )

        # for some reason this makes the above styles start affecting tooltips
        self.app.setStyleSheet(
            """
            QToolTip {
                border: 1px solid #3A3939;
            }
        """
        )

        # if no package data (PackageCollection object) for a houdini install, then there's no way to know where
        # the Houdini /packages directory is, so remove the whole houdini install from the data
        valid_installs = {}
        for version, hou_install in self.houdini_data.hou_installs.items():
            if hou_install.packages:
                valid_installs[version] = hou_install
            else:
                logging.warning(f"Skipping Houdini {version} since no packages were found for it.")
        self.houdini_data.hou_installs = valid_installs

        self.versions = [version.version.front for version in self.houdini_data.hou_installs.values()]

        # MENU BAR
        # menu_bar = self.menuBar()
        # file_menu = menu_bar.addMenu("File")
        # quit_action = file_menu.addAction("Quit")
        # quit_action.triggered.connect(self.quit_app)

        # STATUS BAR
        self.status_bar = QStatusBar(self)
        self.setStatusBar(self.status_bar)
        self.statusLabel = QLabel()
        self.status_bar.addWidget(self.statusLabel)
        self.statusLabel.setTextInteractionFlags(Qt.TextSelectableByMouse)
        self.statusLabel = self.status_bar.findChild(QLabel)

        logo = SvgPushButton(self, BtnSize.HPM_LOGO, BtnIcon.HPM_LOGO, clickable=False)

        # TABS
        self.tabs = QTabWidget()
        tab_packages = QWidget()
        tab_packages.setProperty("id", "package_table")
        tab_add_packages = QWidget()
        tab_add_packages.setProperty("id", "local_plugin_adder")
        self.tabs.addTab(tab_packages, "Packages")
        self.tabs.addTab(tab_add_packages, "Add Local Plugins")

        self.tabs.setStyleSheet(
            """
            QTabWidget::pane {
                background-color: #303030;
            }

            QTabBar::tab {
                background-color: #303030;
                color: lightgrey;
                font-weight: bold;
                padding: 15px 20px;
            }

            QTabBar::tab:hover { background-color: #666666; }
            QTabBar::tab:selected { background-color: #4d4d4d; }
        """
        )

        # SETTINGS BUTTONS
        self.DOWNLOAD_URL = "https://houpm.com"
        self.REPOSITORY_URL = "https://github.com/ariffjeff/houdini-package-manager"
        self.REPOSITORY_ISSUES_URL = "https://github.com/ariffjeff/houdini-package-manager/issues"
        self.PACKAGES_DOCS_URL = "https://www.sidefx.com/docs/houdini/ref/plugins.html"

        button_houpm = SvgPushButton(self, BtnSize.HPM_LOGO, BtnIcon.HPM_LOGO_GREY)
        button_houpm.clicked.connect(lambda: self.open_url(self.DOWNLOAD_URL))
        button_houpm.set_hover_status_message(f"Open project download site: {self.DOWNLOAD_URL}")
        button_houpm.setToolTip("Project download site")

        button_repo = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.REPO)
        button_repo.clicked.connect(lambda: self.open_url(self.REPOSITORY_URL))
        button_repo.set_hover_status_message(f"Open project repository: {self.REPOSITORY_URL}")
        button_repo.setToolTip("Project repository")

        button_bug_report = SvgPushButton(self, BtnSize.SQUARE_DEFAULT, BtnIcon.BUG)
        button_bug_report.clicked.connect(lambda: self.open_url(self.REPOSITORY_ISSUES_URL))
        button_bug_report.set_hover_status_message(f"Open feature request/bug report: {self.REPOSITORY_ISSUES_URL}")
        button_bug_report.setToolTip("Feature request/bug report")

        button_pkg_docs = SvgPushButton(self, BtnSize.DOCS, BtnIcon.DOCS)
        button_pkg_docs.clicked.connect(lambda: self.open_url(self.PACKAGES_DOCS_URL))
        button_pkg_docs.set_hover_status_message(f"Open Houdini packages docs: {self.PACKAGES_DOCS_URL}")
        button_pkg_docs.setToolTip("Houdini packages docs")

        # TAB DATA
        packages = PackagesWidget(self, self.houdini_data, self.versions, self.tabs)
        add_packages = LocalPackageAdderWidget(self, self.houdini_data, self.versions)

        # CREATE LAYOUTS
        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)

        layout_main_vertical = QVBoxLayout()  # includes the tabs
        layout_main_header = QHBoxLayout()
        layout_urls = QHBoxLayout()

        # SET LAYOUTS
        central_widget.setLayout(layout_main_vertical)

        layout_main_vertical.addLayout(layout_main_header)
        layout_main_header.addWidget(logo)
        layout_main_header.addLayout(layout_urls)
        layout_urls.addWidget(button_houpm)
        layout_urls.addWidget(button_repo)
        layout_urls.addWidget(button_bug_report)
        layout_urls.addWidget(button_pkg_docs)
        layout_urls.setAlignment(Qt.AlignRight)
        layout_main_vertical.addWidget(self.tabs)

        tab_packages.setLayout(packages.layout_main)
        tab_add_packages.setLayout(add_packages.layout_main)

    def open_url(self, url: str):
        """
        Open the repository for this project.
        """

        url = QUrl(url)
        QDesktopServices.openUrl(url)
        StatusBar.message(f"Opened: {url}")

    def quit_app(self):
        self.app.quit()



import json
from pathlib import Path
from typing import List

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QAbstractItemView,
    QCheckBox,
    QDialog,
    QFileDialog,
    QFileSystemModel,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QListView,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QTreeView,
    QVBoxLayout,
    QWidget,
)

from houdini_package_manager.meta.meta_tools import StatusBar, TextColor
from houdini_package_manager.widgets.custom_widgets import BtnIcon, BtnSize, SvgPushButton
from houdini_package_manager.wrangle.config_control import HoudiniManager
from houdini_package_manager.wrangle.package_templates import PackageTemplates


class LocalPackageAdderWidget(QWidget):
    """
    The layout that displays the tools for creating new packages from plugin folders/files.

    Arguments:
        parent:
            The parent widget.

        table_data (HoudiniManager):
            The data set to populate the QWidgetTable. Data can consist of multiple sets of packages
            for different installed versions of Houdini.

        versions (List[str]):
            The list of ordered version numbers that will determine which set of package data is shown in the table.
    """

    def __init__(self, parent, table_data: HoudiniManager, versions: List[str]) -> None:
        super().__init__(parent)

        self.table_data = table_data
        self.versions = versions
        self.tool_widgets = PresetWidgets()

        # LIST - HOUDINI VERSIONS
        label_versions = QLabel("Create packages for:")
        button_invert_versions_selection = self.tool_widgets.invert_selection(self)
        button_clear_versions_selection = self.tool_widgets.clear_selection(self)

        self.layout_v_hou_versions = QVBoxLayout()
        self.layout_v_hou_versions.setAlignment(Qt.AlignTop)
        for version in self.versions:
            checkbox = QCheckBox(f"Houdini {version}")
            checkbox.setProperty("version", version)
            checkbox.clicked.connect(self.update_file_overwrite_warning)
            self.layout_v_hou_versions.addWidget(checkbox)
        button_invert_versions_selection.clicked.connect(lambda: self.invert_selection(self.layout_v_hou_versions))
        button_clear_versions_selection.clicked.connect(lambda: self.clear_selection(self.layout_v_hou_versions))

        group_versions = QGroupBox()
        group_versions.setStyleSheet(
            """
            QGroupBox {
                border: 1px solid gray;
                padding: 5px;
            }
        """
        )
        group_versions.setLayout(self.layout_v_hou_versions)

        # LIST - PLUGINS TO ADD
        label_plugins = QLabel("Plugins to add:")
        button_add_items = self.tool_widgets.add_items(self)
        button_invert_plugins_selection = self.tool_widgets.invert_selection(self)
        button_clear_plugins_selection = self.tool_widgets.clear_selection(self)
        remove_selected_items = self.tool_widgets.remove_selected_items(self)
        button_remove_all_items = self.tool_widgets.remove_all_items(self)

        self.label_file_overwrite = QLabel("")
        self.label_file_overwrite.setProperty("update", self.update_file_overwrite_warning)

        self.plugin_list_box = PluginListBox(self, self.label_file_overwrite)
        self.plugin_list_box.setStyleSheet(
            """
            PluginListBox {
                border: 1px solid gray;
                padding: 15px;
            }
        """
        )
        button_add_items.clicked.connect(self.plugin_list_box.add_plugins_to_list)
        button_invert_plugins_selection.clicked.connect(self.plugin_list_box.invert_selection)
        button_clear_plugins_selection.clicked.connect(self.plugin_list_box.clear_selection)
        remove_selected_items.clicked.connect(self.plugin_list_box.remove_selected_items)
        button_remove_all_items.clicked.connect(self.plugin_list_box.remove_all_items)

        # BUTTON - CREATE PACKAGES
        button_add_package = SvgPushButton(self, BtnSize.WIDE_LARGE, BtnIcon.ADD_PKG_CONFIRM)
        button_add_package.set_hover_status_message(
            "Create packages for the selected plugins for the selected Houdini versions."
        )
        button_add_package.setToolTip("Create packages")
        button_add_package.clicked.connect(self.create_packages)

        # CREATE LAYOUTS
        self.layout_main = QVBoxLayout()
        layout_h_versions_and_plugins = QHBoxLayout()

        layout_v_versions = QVBoxLayout()
        layout_h_versions_header = QHBoxLayout()

        layout_v_plugins = QVBoxLayout()
        layout_h_plugins_header = QHBoxLayout()

        # SET LAYOUTS
        self.layout_main.addLayout(layout_h_versions_and_plugins)
        layout_h_versions_and_plugins.addLayout(layout_v_versions, 1)
        layout_h_versions_and_plugins.addLayout(layout_v_plugins, 4)
        self.layout_main.addWidget(button_add_package)

        layout_v_versions.addLayout(layout_h_versions_header)
        layout_h_versions_header.addWidget(label_versions)
        layout_h_versions_header.addWidget(button_invert_versions_selection)
        layout_h_versions_header.addWidget(button_clear_versions_selection)
        layout_h_versions_header.setAlignment(label_versions, Qt.AlignBottom)
        layout_v_versions.addWidget(group_versions)

        layout_v_plugins.addLayout(layout_h_plugins_header)
        layout_h_plugins_header.addWidget(label_plugins)
        layout_h_plugins_header.addWidget(button_add_items)
        layout_h_plugins_header.addWidget(button_invert_plugins_selection)
        layout_h_plugins_header.addWidget(button_clear_plugins_selection)
        layout_h_plugins_header.addWidget(remove_selected_items)
        layout_h_plugins_header.addWidget(button_remove_all_items)
        layout_h_plugins_header.setAlignment(label_plugins, Qt.AlignBottom)
        layout_v_plugins.addWidget(self.plugin_list_box)
        layout_v_plugins.addWidget(self.label_file_overwrite)

    def update_file_overwrite_warning(self) -> None:
        """
        The warning label will update to show the number of potential package config
        overwrites, if any of the plugins' new config files may overwrite existing
        config files in the selected Houdini versions' package directories.
        """

        versions_selected = self._selected_versions()
        plugins_selected = self.plugin_list_box._selected_plugin_items()
        if not versions_selected or not plugins_selected:
            self.label_file_overwrite.setText("")
            return

        plugin_paths = [plugin.path for plugin in plugins_selected]
        packages_directories = [
            self.table_data.hou_installs[version].packages.packages_directory for version in versions_selected
        ]

        # assemble the paths of the new config files
        expected_new_package_paths = []
        for directory in packages_directories:
            for path in plugin_paths:
                expected_new_package_paths.append(Path(directory, path.name).with_suffix(".json"))

        existing_paths = 0
        for path in expected_new_package_paths:
            if path.exists():
                existing_paths += 1

        if existing_paths == 0:
            self.label_file_overwrite.setText("")
            return

        self.label_file_overwrite.setText(f"{existing_paths} existing config files will be overwritten.")

    def _selected_versions(self) -> List[str]:
        """
        Return a list of the selected houdini version numbers from the version list.
        """

        widgets = []
        for i in range(self.layout_v_hou_versions.count()):
            widgets.append(self.layout_v_hou_versions.itemAt(i).widget())

        widgets = [widget.property("version") for widget in widgets if widget.isChecked()]

        return widgets

    def _create_config_data(self, plugin_paths: List[Path]) -> dict:
        """
        Create the structured config data from a package template for each plugin path.
        """

        new_packages = {}
        for path in plugin_paths:
            new_packages[path.name] = PackageTemplates.standard(path)

        return new_packages

    def create_packages(self) -> None:
        """
        Create the package configs in the relevant Houdini packages
        directories based on the selected Houdini versions and the selected
        plugin items.
        """

        loaded = self.plugin_list_box._loaded_items()
        if not loaded:
            StatusBar.message("No plugins in list.", TextColor.ERROR)
            return

        plugins = self.plugin_list_box._selected_plugin_items()
        if not plugins:
            StatusBar.message("No plugins selected.", TextColor.ERROR)
            return

        versions = self._selected_versions()
        if not versions:
            StatusBar.message("No Houdini versions selected.", TextColor.ERROR)
            return

        plugin_paths = [plugin.path for plugin in plugins]
        packages_directories = [
            self.table_data.hou_installs[version].packages.packages_directory for version in versions
        ]

        new_packages = self._create_config_data(plugin_paths)

        for directory in packages_directories:
            for name, package in new_packages.items():
                outpath = Path(directory, name).with_suffix(".json")
                with open(outpath, "w") as outfile:
                    json.dump(package, outfile, indent=4)

        StatusBar.message(
            f"Created {len(new_packages) * len(packages_directories)} total package configs for"
            f" {len(packages_directories)} Houdini versions.",
            TextColor.SUCCESS,
        )

        self.label_file_overwrite.property("update")()

    def clear_selection(self, layout) -> None:
        """
        Deselects all houdini version list items.
        """

        for i in range(layout.count()):
            box = layout.itemAt(i).widget()
            if isinstance(box, QCheckBox):
                box.setChecked(False)

        self.label_file_overwrite.property("update")()

    def invert_selection(self, layout) -> None:
        """
        Invert the selection of houdini version list items.
        """

        for i in range(layout.count()):
            box = layout.itemAt(i).widget()
            if isinstance(box, QCheckBox):
                box.setChecked(not box.isChecked())

        self.label_file_overwrite.property("update")()


class PluginListBox(QStackedWidget):
    def __init__(self, parent, label_file_overwrite: QLabel):
        super().__init__(parent)

        self.label_file_overwrite = label_file_overwrite
        # the system paths selected by the user from the dialog box
        self.selected_paths = []

        # the default "add plugin folders" button widget
        button_open = SelectPluginsButton(self, "Select plugin folders...")
        button_open.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        button_open.clicked.connect(self.add_plugins_to_list)

        widget_add_plugins = QWidget()
        layout_add_plugins = QVBoxLayout(widget_add_plugins)
        layout_add_plugins.setContentsMargins(0, 0, 0, 0)
        layout_add_plugins.addWidget(button_open)

        # the widget layout of plugin path items
        widget_plugin_list = QWidget()
        self.layout_plugin_list = QVBoxLayout(widget_plugin_list)
        self.layout_plugin_list.setContentsMargins(0, 0, 0, 0)
        self.layout_plugin_list.setAlignment(Qt.AlignTop)
        # plugin item widgets get added later by user via dialog choice

        # stack layout
        self.addWidget(widget_add_plugins)
        self.addWidget(widget_plugin_list)
        self.setCurrentIndex(0)

        self.setStyleSheet("QStackedWidget { border: 1px solid grey; }")

    def remove_all_items(self) -> None:
        """
        Remove all items from the plugin list layout.
        """

        while self.layout_plugin_list.count():
            item = self.layout_plugin_list.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.setParent(None)
            else:
                self.layout_plugin_list.removeItem(item)

        self.setCurrentIndex(0)
        self.label_file_overwrite.property("update")()

    def remove_selected_items(self) -> None:
        """
        Remove only the selected items from the plugin list layout.
        """

        checked_boxes = self._selected_plugin_items()
        for widget in checked_boxes:
            self.layout_plugin_list.removeWidget(widget)
            widget.deleteLater()

        # if no items left, change stack back to default
        if self.layout_plugin_list.count() == 0:
            self.setCurrentIndex(0)
        self.label_file_overwrite.property("update")()

    def invert_selection(self) -> None:
        """
        Invert the selection of plugin list items.
        """

        items = self._loaded_items()
        for item in items:
            item.checkbox.toggle()
        self.label_file_overwrite.property("update")()

    def clear_selection(self) -> None:
        """
        Deselects all plugin list items.
        """

        items = self._loaded_items()
        for item in items:
            item.checkbox.setChecked(False)
        self.label_file_overwrite.property("update")()

    def _selected_plugin_items(self) -> list:
        """
        Return list of plugin item widgets that are selected (checked checkboxes).
        """

        loaded_widgets = self._loaded_items()
        checked_boxes = []
        for widget in loaded_widgets:
            if widget.checkbox.isChecked():
                checked_boxes.append(widget)

        return checked_boxes

    def _loaded_items(self) -> list:
        """
        Return a list of all the currently loaded widgets in the plugin list layout.
        """

        widget_list = []
        for i in range(self.layout_plugin_list.count()):
            widget = self.layout_plugin_list.itemAt(i).widget()
            if widget is not None:
                widget_list.append(widget)

        return widget_list

    def add_plugins_to_list(self) -> None:
        """
        Get the user selected system paths, convert them into custom
        list items for the plugins box layout, and switch to the plugin
        list layout in the stacked widget.

        If there are existing plugin items in the list layout,
        they will be appended onto.

        Paths already in the plugin list layout are not duplicated if
        the user chooses the same path again from the dialog box.
        """

        self._open_files_and_folders()

        if not self.selected_paths:
            return

        # ignore duplicate paths
        loaded_paths = [widget.path for widget in self._loaded_items()]
        paths_to_add = [Path(path) for path in self.selected_paths if Path(path) not in loaded_paths]
        for path in paths_to_add:
            self.layout_plugin_list.addWidget(PluginListItem(self, path, self.label_file_overwrite))

        self.setCurrentIndex(1)
        self.selected_paths = []

        self.label_file_overwrite.property("update")()

    def _open_files_and_folders(self) -> None:
        """
        This is the absolute best attempt to get multiple file & folder selection working
        in one dialog box. It won't work using a native system dialog box, it only seems
        to work using QFileDialog.Option.DontUseNativeDialog.

        The downside is that the user can't paste in a directory path to jump anywhere in the
        dialog and has to manually navigate the top dropdown bar.

        The other downside is that to select any file, a folder must also be selected, and the
        folder must be the first item selected in order for the dialog box to be accepted.

        Selected paths are converted to pathlib Paths.
        """

        dialog = QFileDialog(self)
        dialog.setStyleSheet(
            """
            QHeaderView::section
            {
                background-color: #333;
                color: white;
            }

            QPushButton:disabled {
                color: #444;
            }

            * { font-size: 14px; }
        """
        )
        dialog.setWindowTitle("Select plugins folders/files:")
        dialog.setOption(QFileDialog.Option.DontUseNativeDialog, True)
        dialog.setFileMode(QFileDialog.FileMode.Directory)

        for view in dialog.findChildren(QListView) + dialog.findChildren(QTreeView):
            if isinstance(view.model(), QFileSystemModel):
                view.setSelectionMode(QAbstractItemView.ExtendedSelection)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            self.selected_paths = dialog.selectedFiles()
        dialog.deleteLater()

    def _open_folder(self) -> None:
        # single folder selection
        folderpath = QFileDialog.getExistingDirectory(
            self,
            "Select plugin folder:",
            "",
            # "HDAs/OTLs (*.hda *.hdalc *.hdanc *.hda*);;All Files (*)",
        )
        self.selected_paths = folderpath


class PluginListItem(QWidget):
    """
    A plugin row for the plugins list.
    Includes a checkbox (with the path name), the path, and a remove-item button.
    """

    def __init__(self, listbox: PluginListBox, path: Path, label_file_overwrite: QLabel) -> None:
        super().__init__(listbox)

        if not isinstance(path, Path):
            raise TypeError("path must be a pathlib Path.")

        self.listbox = listbox
        self.label_file_overwrite = label_file_overwrite

        self.button_remove = SvgPushButton(self, BtnSize.SQUARE_SMALL, BtnIcon.REMOVE)
        self.button_remove.set_hover_status_message("Remove item.")
        self.button_remove.setToolTip("Remove item")
        self.button_remove.clicked.connect(self.remove_self)

        self.path = path
        self.name = path.name
        self.checkbox = QCheckBox(path.name)
        self.checkbox.setChecked(True)
        self.checkbox.clicked.connect(self.toggled)
        self.label = QLabel(str(path))
        self.label.setStyleSheet("color: grey;")

        layout_main = QHBoxLayout()
        layout_main.addWidget(self.button_remove)
        layout_main.addWidget(self.checkbox)
        layout_main.addWidget(self.label)
        self.setLayout(layout_main)
        layout_main.setContentsMargins(0, 0, 0, 0)

    def toggled(self):
        """
        Events that occur when the QCheckBox is clicked.
        """

        self.label_file_overwrite.property("update")()

    def remove_self(self):
        """
        Deletes this widget.
        """

        self.deleteLater()

        # since the delete happens later and we need to update the file overwrite conflict check now,
        # manually disable the toggle instead, which has the same effect in this case.
        self.checkbox.setChecked(False)
        self.label_file_overwrite.property("update")()

        # if there's only one item left, then it must be this one (self) that's getting deleted,
        # so change stack back to default
        if self.listbox.layout_plugin_list.count() == 1:
            self.listbox.setCurrentIndex(0)


class SelectPluginsButton(QPushButton):
    """
    The default button that prompts the user to open a file/folder dialog and select files/folders.
    This is meant to be displayed when the plugins list box has no items yet.
    """

    def __init__(self, parent, text):
        super().__init__(parent=parent, text=text)

        self.setStyleSheet("border: none;")

    def enterEvent(self, event):
        self.setCursor(Qt.PointingHandCursor)

    def leaveEvent(self, event):
        self.setCursor(Qt.ArrowCursor)


class PresetWidgets:
    """
    Tool widgets for the package adder window.
    """

    @classmethod
    def add_items(cls, parent) -> SvgPushButton:
        add_items = SvgPushButton(parent, BtnSize.WIDE_DEFAULT, BtnIcon.ADD)
        add_items.set_hover_status_message("Select plugins.")
        add_items.setToolTip("Select plugins")
        return add_items

    @classmethod
    def invert_selection(cls, parent) -> SvgPushButton:
        invert_selection = SvgPushButton(parent, BtnSize.SQUARE_MEDIUM, BtnIcon.INVERT_SELECTION)
        invert_selection.set_hover_status_message("Invert selection.")
        invert_selection.setToolTip("Invert selection")
        return invert_selection

    @classmethod
    def clear_selection(cls, parent) -> SvgPushButton:
        clear_selection = SvgPushButton(parent, BtnSize.SQUARE_MEDIUM, BtnIcon.CLEAR_SELECTION)
        clear_selection.set_hover_status_message("Clear selection.")
        clear_selection.setToolTip("Clear selection")
        return clear_selection

    @classmethod
    def remove_selected_items(cls, parent) -> SvgPushButton:
        remove_item = SvgPushButton(parent, BtnSize.SQUARE_MEDIUM, BtnIcon.REMOVE)
        remove_item.set_hover_status_message("Remove selected items.")
        remove_item.setToolTip("Remove selected items")
        return remove_item

    @classmethod
    def remove_all_items(cls, parent) -> SvgPushButton:
        remove_all_items = SvgPushButton(parent, BtnSize.SQUARE_MEDIUM, BtnIcon.REMOVE_ALL_ITEMS)
        remove_all_items.set_hover_status_message("Remove all items.")
        remove_all_items.setToolTip("Remove all items")
        return remove_all_items



import logging

import requests
from packaging import version
from PySide6.QtCore import Qt, QUrl
from PySide6.QtGui import QDesktopServices
from PySide6.QtWidgets import QDialog, QHBoxLayout, QLabel, QMessageBox, QPushButton, QVBoxLayout, QWidget

from houdini_package_manager import __version__


class Updater(QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)

    def check_update(self) -> None:
        """
        Check for the latest version of the installed package on PyPI.
        Prompt upgrade if installed app is out of date.
        """

        self.dist_url = QUrl("https://houpm.com")
        self.TITLE = "houdini-package-manager"

        latest_version = self.get_latest_project_version()

        if not latest_version:
            return

        if version.parse(latest_version) > version.parse(__version__):
            update_dialog = UpdateDialog(self, latest_version, self.dist_url)
            update_dialog.setStyleSheet(
                """
                background-color: #303030;
                color: white;
                font-size: 14px;
            """
            )

            choice = update_dialog.exec()
            if choice == QMessageBox.Yes:
                QDesktopServices.openUrl(self.dist_url)

    def get_latest_project_version(self) -> str:
        """
        Get the latest version of the project from PyPI.
        Returns None if the version number can not be retrieved.
        """

        try:
            logging.debug("Checking for HPM updates...")
            response = requests.get(f"https://pypi.org/pypi/{self.TITLE}/json", timeout=5)
        except (requests.RequestException, ValueError):
            return None

        if response.status_code == 200:
            data = response.json()
            latest_version = data["info"]["version"]
            return latest_version
        return None


class UpdateDialog(QDialog):
    """
    The dialog that displays the new version project version information
    and the choice to update or not.
    """

    def __init__(self, parent, latest_version: str, dist_url: QUrl):
        super().__init__(parent)

        self.latest_version = latest_version
        self.dist_url = dist_url

        self.setWindowTitle("Upgrade Dialog")

        layout = QVBoxLayout(self)

        label = QLabel(f"A new version ({latest_version}) of HPM is available.\n\nDo you want to upgrade?")
        label.setAlignment(Qt.AlignLeft)
        layout.addWidget(label)

        button_layout = QHBoxLayout()
        button_layout.setContentsMargins(0, 20, 0, 0)
        layout.addLayout(button_layout)

        yes_button = QPushButton("Yes")
        yes_button.clicked.connect(self.upgrade)
        button_layout.addWidget(yes_button)

        no_button = QPushButton("No")
        no_button.clicked.connect(self.reject)
        button_layout.addWidget(no_button)

    def upgrade(self) -> None:
        QDesktopServices.openUrl(self.dist_url)
        self.accept()


class ErrorDialog(QDialog):
    """
    A standard error dialog that contains a message and an OK button.
    """

    def __init__(self, error_message):
        super().__init__()
        self.setWindowTitle("Error: Houdini not installed")
        self.setFixedWidth(300)
        self.setFixedHeight(150)
        self.layout_main = QVBoxLayout()
        self.setLayout(self.layout_main)

        label = QLabel(error_message)
        label.setWordWrap(True)
        self.layout_main.addWidget(label)

        ok_button = QPushButton("OK")
        ok_button.setMaximumWidth(85)
        ok_button.clicked.connect(self.accept)
        self.layout_main.addWidget(ok_button)

        self.setStyleSheet(
            """
            background-color: #303030;
            color: white;
            padding: 5px;
        """
        )



import json
import logging
import os
from enum import Enum
from pathlib import Path

from PySide6.QtWidgets import QApplication, QMainWindow, QStatusBar


class TextColor(Enum):
    """
    Text colors determined by CSS style text color.
    """

    DEFAULT = "color: white;"
    ERROR = "color: red;"
    SUCCESS = "color: green;"
    WARNING = "color: yellow;"


class StatusBar:
    """
    The status bar of the application.
    This is a wrapper around QStatusBar.
    """

    @classmethod
    def message(cls, message: str, text_color: TextColor = None) -> None:
        """
        Print a message to the main window's status bar.

        Arguments:
            message (str):
                The string to print to the status bar.

            text_color (TextColor):
                The text color enum that sets the color the status bar message text.
                Default is white.
        """

        if not text_color:
            text_color = TextColor.DEFAULT

        status_bar = cls.status_bar()
        status_bar.setStyleSheet(text_color.value)
        status_bar.showMessage(message)

    @staticmethod
    def status_bar(raise_on_error=True) -> QStatusBar:
        """
        Return the status bar object.
        Raises an error or returns False if the status bar cannot be found depending on what
        the error argument specifies.

        Arguments:
            raise_on_error (bool):
                Whether or not you want to raise an error if the status bar cannot be found.
                This can be useful if you just want to know if the statusbar can be accessed.
                Default is True.
                If False, False will be returned instead.
        """

        top_level_widgets = QApplication.topLevelWidgets()
        for widget in top_level_widgets:
            if isinstance(widget, QMainWindow):
                return widget.statusBar()

        if not raise_on_error:
            return False

        raise RuntimeError("Unable to find status bar in top level widget.")


class TableHeaders(Enum):
    """
    Enums for the table column header names.
    """

    ENABLE = "Enable"
    PACKAGE = "Package"
    AUTHOR = "Author"
    LATEST = "Latest"
    INSTALLED = "Installed"
    SOURCE = "Src"
    CONFIG = "Config"
    PLUGINS = "Plugins"
    SYNC = "Sync"
    UPDATE = "Update"


class RateLimitError(Exception):
    """
    An exception that should be raised when an API's rate limit has been exceeded (status code: 403).
    """

    def __init__(self, message="API rate limit exceeded. Status code: 403") -> None:
        self.message = message
        super().__init__(self.message)


class RequestConnectionError(Exception):
    """
    An exception when a request connection fails.
    """

    def __init__(self, message="Failed to establish connection.") -> None:
        self.message = message
        super().__init__(self.message)


class UserDataManager:
    """
    TODO
    - create new pkg name dict in json if it doesnt exist (or just pre init them all at once?)
    - read values on HPM load
    - if a non existent json entry for a tool is created, init its local_config_path (forgot to do that currently)
    - account for json file not existing
    - read data back in on HPM start
    - namespace each tool in json user data to be owner.tool to prevent entry collisions
    - ensure that both owner name and tool name variables used to write/read json are identical across config_control and anywhere else.
    - use 'from dataclasses import dataclass, field' to maintain structured data when reading/writing json
    """

    def __init__(self):
        self.file_path = Path("houdini_package_manager/user/package_repo_data.json")

    def _read_data(self) -> dict:
        """
        Reads data from the JSON file.

        If the file doesn't exist, it will be created with no data.
        """

        if self.file_path.exists():
            with open(self.file_path) as file:
                return json.load(file)
        else:
            self.new_empty_file()
            return {}

    def _write_data(self, data) -> None:
        """Writes the given data to the JSON file."""
        with open(self.file_path, "w") as file:
            json.dump(data, file, indent=4)

    # def add_entry(self, tool_name, local_config_path):
    #     """Adds a new entry to the data."""
    #     data = self._read_data()
    #     data[tool_name] = {"local_config_path": local_config_path, "tags": []}
    #     self._write_data(data)

    def update_tags(self, tool_name, tags) -> None:
        """Updates the tags for a specific tool."""
        data = self._read_data()
        if tool_name not in data:
            # If the tool does not exist, initialize its entry with empty tags
            data[tool_name] = {"local_config_path": "", "tags": []}
        data[tool_name]["tags"] = tags
        self._write_data(data)

    def get_entry(self, tool_name) -> dict | None:
        """Retrieves the entry for a specific tool."""
        if not tool_name:
            return

        data = self._read_data()
        if tool_name in data:
            return data[tool_name]
        else:
            logging.debug(f"User data cache for plugin '{tool_name}' does not exist.")
            return None

    def set_file_path(self, file_path):
        """Sets or changes the file path for the JSON data file."""
        self.file_path = file_path

    def new_empty_file(self) -> None:
        """
        Creates the missing file with no data.

        If the user folder does not exist, it will also be created first.
        """

        folder_path = os.path.dirname(self.file_path)
        os.makedirs(folder_path, exist_ok=True)  # ensure the directory exists

        with open(self.file_path, "w") as file:
            json.dump({}, file)


class SingletonMeta(type):
    """
    A metaclass to support other classes to make them singletons.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


class GlobalExceptionTracker(metaclass=SingletonMeta):
    """
    Singleton global class for tracking the latest raised exception.

    This was written to work around the issue of exceptions that get raised by
    a Qt button's emitted signal (set via a slot) but fail to get passed to the
    caller when the button's signals were called programmatically with something like
    `some_qpushbutton.clicked.emit()` instead of being clicked in the UI. This meant
    that exceptions would get printed to the terminal/debugger but not trigger the
    exception part of a try-catch block.

    This class allows for keeping track of an exception if one is created and
    then later checking for its existence where try-catch block would usually be.
    It can effectively be used to mimic a try-catch.

    A single exception must be written, read, and cleared on this object manually
    with the relevant methods.
    """

    def __init__(self):
        self._exception = None

    def set_exception(self, exception) -> None:
        """
        Sets the given exception as the most recent exception.
        """

        if not issubclass(type(exception), Exception):
            raise TypeError(f"Argument {exception} must be a subclass of Exception.")

        self.clear_exception()
        self._exception = exception

    def get_exception(self) -> Exception:
        """
        Returns the exception that was last set.

        The exception can be of any type with a base class of Exception.
        """

        exception = self._exception
        self.clear_exception()
        return exception

    def clear_exception(self) -> None:
        """
        Clears the last set exception by setting the attribute to None.
        """

        self._exception = None



from urllib.parse import urlparse


class Url:
    """
    A URL path and its parsed result.
    """

    def __init__(self, url: str | None) -> None:
        if not isinstance(url, str):
            raise TypeError("URL must be a string.")

        self._url = url
        self.parse = urlparse(url)  # ParseResult

    def __str__(self) -> str:
        return self._url

    def __repr__(self) -> str:
        return self._url

    @property
    def stem(self) -> str:
        """
        The final path element of the URL, excluding the extension.
        """

        s = self.__str__()
        s = s.split("/")
        stem = s[-1]
        stem = stem.split(".")
        stem = stem[0]
        return stem



from __future__ import annotations

import json
import logging
import os
import platform
import re
import subprocess
from itertools import takewhile
from pathlib import Path

from houdini_package_manager.meta.meta_tools import TableHeaders
from houdini_package_manager.wrangle.repository import GitProject
from houdini_package_manager.wrangle.url import Url


class HoudiniManager:
    """
    A class for managing data related to multiple installed versions of Houdini.

    Arguments:
        only_hou_installs (bool):
            Remove any paths that aren't Houdini install paths.
            This will remove the LicenceServer path for example.

    Attributes:
        install_directories (Dict[str, Path]):
            The directories containing the install locations of different Houdini versions.
            The keys are the version numbers. The values are the directories.

        hou_installs (Dict[str, HoudiniInstall]):
            A dictionary of all the installed versions of Houdini and their data.
    """

    def __init__(self, only_hou_installs=True) -> None:
        self.install_directories = self._get_houdini_paths()
        if only_hou_installs:
            self.install_directories = self.only_houdini_locations()

        logging.debug(f"Using these installs:\n{self.install_directories}\n")
        self.hou_installs = {}

    def get_houdini_data(self, versions: str | list[str] = None) -> None:
        """
        Get the package data and relevant meta data for each installed version of Houdini.
        If an install has no package data then the config will simply be empty.

        Arguments:
            version (str | List[str]):
                Get all data for only the given Houdini version(s). If the data already exists
                for a version then it is replaced by a new set of data.
        """

        if isinstance(versions, str):
            versions = [versions]

        # replace only the data for the given versions
        if versions:
            for ver in versions:
                try:
                    self.hou_installs[ver] = HoudiniInstall(self.install_directories[ver])
                except KeyError as e:
                    raise KeyError(
                        f"Houdini version {e} is not in the dict of known Houdini install directories."
                    ) from e
            return

        for ver, path in self.install_directories.items():
            self.hou_installs[ver] = HoudiniInstall(path)

    def _get_houdini_paths(self) -> dict:
        """
        Get the locations of Houdini installed software and related software, independent of OS.
        This includes any Houdini version and other components like LicenseServer.
        """

        opsys = platform.system()

        paths = {}
        if opsys == "Windows":
            paths = self._win_registry_values(r"SOFTWARE\Side Effects Software\Houdini")
        elif opsys == "Darwin" or opsys == "Linux":
            pass
        else:
            raise Exception("Could not determine operating system.")

        paths = {key: path for key, path in paths.items() if path.exists()}

        # only get houdini version paths
        pattern = re.compile(r"^\d+\.\d+")  # match version numbers (19.0 & 19.5.123)
        paths = {k: v for k, v in paths.items() if pattern.match(k)}

        # sort dict items by key version number in descending order (largest version number first)
        paths = dict(sorted(paths.items(), key=lambda x: tuple(map(int, x[0].split("."))), reverse=True))
        logging.debug(f"Houdini install paths:\n{paths}\n")
        return paths

    def _win_registry_values(self, key_path: str) -> dict:
        """
        Get the values of a Windows registry key.
        Paths are converted to pathlib.Path objects.
        """

        import winreg

        try:
            # Open the registry key
            key = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key_path)

            # Iterate over all values in the key
            values = {}
            i = 0
            while True:
                # Get the name, data, and type of the next value
                name, data, _ = winreg.EnumValue(key, i)
                name = self._houdini_version_name(name)
                if isinstance(data, str):
                    data = Path(data)
                values[name] = data
                i += 1
        except OSError:
            # no more values left
            pass

        logging.debug(f"Houdini relevant registry keys detected:\n{values}\n")
        return values

    def _houdini_version_name(self, version: str) -> str:
        """
        Get a combination of the major, minor, and patch version components from the full semantic version number
        of an installed version of Houdini.
        Removes the mystery 3rd middle version component that doesn't seem to matter or show up anywhere else.
        """

        # which components of the version number to use.
        # major, minor, mystery number, patch
        version_components = [True, True, False, False]

        if any(char.isalpha() for char in version):
            return version

        version = version.split(".")
        new_version = [item for item, bool_val in zip(version, version_components) if bool_val]
        new_version = ".".join(new_version)

        return new_version

    def only_houdini_locations(self) -> dict:
        """
        Remove paths that aren't Houdini install locations.
        e.g. This is a Houdini install directory:
            'C:/Program Files/Side Effects Software/Houdini 19.0.917'
        """

        keys_to_remove = []
        for key in self.install_directories:
            key_split = key.split(".")
            try:
                int(key_split[0])
            except Exception:
                keys_to_remove.append(key)

        install_dirs_copy = dict(self.install_directories)
        for key in keys_to_remove:
            del install_dirs_copy[key]

        return install_dirs_copy


class HoudiniInstall:
    """
    A manager of all the relevant data for a single installed version of Houdini.
    """

    def __init__(self, install_dir: Path) -> None:
        if install_dir and not isinstance(install_dir, Path):
            raise TypeError("install_dir must be a pathlib.Path object.")

        self.HFS = install_dir
        self.HB = Path(self.HFS, "bin")
        self.version = HouVersion(str(self.HFS))
        self.env_vars = self._get_env_vars()
        self.packages = self._get_pkgs()
        self._final_debug_logs()

    def _final_debug_logs(self) -> None:
        """
        Various debug logs to explain the results of this Houdini version's extracted package/plugin data.
        """

        if not self.packages:
            return

        logging.debug(f"Houdini {self.version.full} PACKAGE CONFIGS:")
        for pkg in self.packages.pkgs.values():
            logging.debug(pkg.config_path)
        logging.debug("\n")

        logging.debug(f"Houdini {self.version.full} PLUGINS:")
        for plugin in self.packages.hconfig_plugin_paths:
            logging.debug(plugin)
        logging.debug("\n")

    def _get_pkgs(self) -> PackageCollection | None:
        """
        Returns the PackageCollection object for this Houdini version.
        Returns None if the Houdini user preferences directory does not exist as an environment variable.
        """

        HOU_PKG_DIR_KEY = "HOUDINI_USER_PREF_DIR"
        # if hconfig.exe failed to produce the "HOUDINI_USER_PREF_DIR" env var or any env var data
        # because maybe the houdini install is corrupted somehow, then no package data can be retrieved.
        logging.debug(f"Houdini {self.version.full} ENV VARS:")
        for key, value in self.env_vars.items():
            logging.debug(f"{key} = {value}")
        logging.debug("\n")

        if HOU_PKG_DIR_KEY in self.env_vars:
            return PackageCollection(Path(self.env_vars[HOU_PKG_DIR_KEY], "packages"), self.env_vars)

        logging.warning(
            f"'{HOU_PKG_DIR_KEY}' was missing from Houdini {self.version.front}'s environment variables.\nThis means"
            " that hconfig.exe probably failed to return it, thus any of this Houdini version's installed packages are"
            " not able to be found."
        )
        return None

    def _get_env_vars(self) -> list:
        """
        Executes Houdini's hconfig.exe via a Python subprocess in order to get the generated Houdini environment variables (keys and values) that
        hconfig processes from the json package config files.

        hconfig.exe can only be run by a compatible Python version that Houdini shipped with (Windows only: https://www.sidefx.com/docs/houdini/hom/index.html#which-python).
        So here we can either:
            1. Find and run a compatible Python by proxy in order to have it run hconfig.exe. This allows us to avoid needing to ship this project with multiple matching Python versions.
            or...
            2. Naively run any Houdini hconfig.exe's with whataver version of Python HPM is built with. For example, a Python3.9 subprocess calling an hconfig.exe that was build for
            Python3.10 will result in an error message being returned, and thus no env vars. No env vars means we can't process any package config data - the given Houdini version
            would have to be ignored entirely! If HPM is running hconfig naively, HPM's Python will need to be updated to match the latest Houdini's Python version on every major release.

        ### WARNING
        Due to the difficulty of getting Houdini environment variables (keys AND values) quickly and without errors, here are the
        different ways to go about it (hconfig and other methods).

        1. open houdini directly > Python Shell > hou.ui.packageInfo()
        - Returns plugin paths (some env var values and no keys), and other crap, can probably be extracted with regex
        - Only works directly in Houdini's python shell

        2. hython > hou.getenv("HOUDINI_PATH")
        - Returns plugin paths (not env vars!)

        3. python > subprocess call hython > import os; os.environ
        - Returns env var keys + values, including global Windows env vars.
        - Slow performance overhead to load hython

        4. python > subprocess > hconfig
        - Returns env var keys + values
        - fast performance
        - errors on incompatible python versions calling hconfig.exe (if this can be fixed then this is the ideal solution)

        5. Somehow find env vars a different way?
        ###

        Returns a list of the Houdini environment variables.
        """

        def _run_with_this_apps_python_naively(hconfig_path: Path) -> list[str]:
            # naively run given hconfig with this project's python version, which might be incompatible (would return useless data)

            subproc_return = subprocess.run(
                [hconfig_path], cwd=hconfig_path.parent, shell=True, capture_output=True, text=True
            )
            subproc_stdout = subproc_return.stdout

            if subproc_stdout:
                logging.debug(f"HCONFIG RETURNED:\n{subproc_stdout}\n")
            else:
                logging.error("HCONFIG FAILED TO RETURN ANY DATA!...")
                logging.error("Subprocess to hconfig returned:")
                logging.error(f"args: {subproc_return.args}")
                logging.error(f"returncode: {subproc_return.returncode}")
                logging.error(f"stderr: {subproc_return.stderr}\n")

            metadata = subproc_stdout.split("\n")
            return metadata

        def _run_with_compatible_python(hconfig_path: Path, python_exe_path: Path) -> list[str]:
            # run hconfig based on the python version it was built for through two-layered subprocess calls.

            # THIS CURRENTLY DOES NOT WORK AS HCONFIG WILL STILL RETURN AN ERROR THINKING IT'S BEING CALLED BY THE
            # WRONG PYTHON VERSION, WHICH MAKES NO SENSE.

            command_main = [
                python_exe_path,
                "-c",
                (
                    f"import subprocess; result = subprocess.run(['{hconfig_path.as_posix()}'], capture_output=True,"
                    " text=True); print(result)"
                ),
            ]
            result = subprocess.run(
                command_main, cwd=hconfig_path.parent, shell=False, capture_output=True, text=True
            ).stdout

            if result:
                logging.debug(f"HCONFIG RETURNED:\n{result}\n")
            else:
                logging.error("HCONFIG FAILED TO RETURN ANY DATA!\n")

            result = re.findall(r'stdout="(.*?)"\s*,\s*stderr=', result)[0]  # extract inner subprocess return value
            metadata = result.split("\\n")
            return metadata

        logging.debug(
            f"Getting Houdini {self.version.full} install data (env vars from hconfig, package data from json)...\n"
        )

        # target the correct hconfig.exe
        hconfig = "hconfig"
        if platform.system() == "Windows":
            hconfig += ".exe"
        hconfig = Path(self.HB, hconfig)

        # run hconfig - choose the best method!
        metadata = _run_with_this_apps_python_naively(hconfig)
        # metadata = _run_with_compatible_python(hconfig, self.this_houdini_python_version())

        metadata = dict(item.split(" := ") for item in metadata if len(item) > 0)
        for key, value in metadata.items():  # remove first and last quotes
            if value[0] in ["'", '"'] and value[-1] in ["'", '"']:
                metadata[key] = value[1:-1]

        return metadata

    def this_houdini_python_version(self) -> Path | None:
        """
        Finds the latest installed version of Python that shipped with this Houdini (Windows only).
        Returns a Path object of the python directory.
        """

        hou_folders = os.listdir(self.HFS)
        python_folders = [folder for folder in hou_folders if re.match(r"python\d+", folder)]

        if not python_folders:
            return None

        installed_pythons = [
            Path(self.HFS, folder, "python.exe")
            for folder in python_folders
            if Path(self.HFS, folder, "python.exe").exists()
        ]
        return installed_pythons[0]

    def pkg_data_as_table_model(self, named=True) -> dict | list:
        """
        Get a list or dict of data for all the packages for a single version of Houdini, ordered as defined by the table model.

        Arguments:
            named (bool):
                Whether or not the returned data should be a dict with the data names as keys or a list of just the values.
                Default is True.
        """
        if named:
            data = {}
            for name, pkg in self.packages.pkgs.items():
                data[name] = pkg.table_model
        else:
            data = []
            for _name, pkg in self.packages.pkgs.items():
                data.append(list(pkg.table_model.values()))
        return data

    def get_package_warnings(self) -> list[str]:
        """
        Get the warnings for each package.
        """

        data = {}
        for name, pkg in self.packages.pkgs.items():
            data[name] = pkg.warnings
        return data

    def get_labels(self) -> list[str]:
        for _, pkg in self.packages.pkgs.items():
            return list(pkg.table_model.keys())


class HouVersion:
    """
    Version components of a Houdini version.
    """

    def __init__(self, install_path: str) -> None:
        if install_path and not isinstance(install_path, str):
            raise TypeError("install_path must be a str.")

        self.full = self._extract_version(install_path)
        parts = self.full.split(".")
        self.major = parts[0]
        self.minor = parts[1]
        self.patch = parts[2]
        self.front = self.major + "." + self.minor

    def _extract_version(self, install_path) -> str:
        """
        Extract the version from the Houdini install path ($HFS).
        """
        ver = os.path.basename(install_path)
        ver = ver.split(" ")[1]
        return ver


class PackageCollection:
    """
    Package JSON configurations and the associated plugin data the configs point to for a given Houdini version.

    Arguments:
        packages_directory (str):
            The directory containing the JSON packages that Houdini references to find plugins.

        env_vars (Dict[str, str]):
            The environment variables needed to help resolve variables found in package configuration
            files. They are aggregated by hconfig

        get_data (bool):
            Whether or not to automatically get the package data and its respective plugin data upon
            creation of the Packages object.
            Default is True.

    Attributes:
        hconfig_plugin_paths (list):
            The paths from HOUDINI_PATH which is from hconfig. These are all
            the plugins paths from all the packages that hconfig found for a single installed
            version of Houdini.
    """

    def __init__(self, packages_directory: Path = None, env_vars: dict[str, str] = None, get_data=True) -> None:
        if packages_directory and not isinstance(packages_directory, Path):
            raise TypeError("directory must be a pathlib.Path object.")

        if not env_vars:
            env_vars = {}
        if not isinstance(env_vars, dict):
            raise TypeError("env_vars must be a dict.")

        if get_data and not isinstance(get_data, bool):
            raise TypeError("get_data must be a bool.")

        self.packages_directory = packages_directory
        self.env_vars = env_vars
        self.hconfig_plugin_paths = []
        self.pkgs = {}  # Package objects
        self.package_plugin_matches = {}

        self.PACKAGES_GIT_DATA_PATH = Path("houdini_package_manager/user_data/package_repo_data.json")

        if get_data:
            self.create_pkgs()

    @property
    def parent_directory(self) -> Path | None:
        if not self.packages_directory:
            return None
        return self.packages_directory.parent

    @property
    def houdini_version(self) -> str:
        """
        The version of Houdini in major.minor format.

        This verion number is extracted directly from the Houdini install directory path.

        Returns a string.
        """

        if not self.packages_directory:
            return None
        return self.packages_directory.parent.name

    def create_pkgs(self) -> None:
        """
        Finds all the package configuration files for a version of Houdini and matches them to their respective plugin data.

        Returns:
            None
        """

        if not self.packages_directory:
            raise AttributeError(
                "directory is not set to any path. Make sure env_vars contains data as well if needed."
            )

        if not self.packages_directory.exists():
            self.packages_directory.mkdir(parents=True)
            logging.debug(f"Created missing packages folder: {self.packages_directory}")

        # add the package directory as a needed environment variable
        # which isn't automatically added by hconfig for some reason.
        HOUDINI_PACKAGE_PATH = "HOUDINI_PACKAGE_PATH"
        if HOUDINI_PACKAGE_PATH not in self.env_vars:
            self.env_vars.update({HOUDINI_PACKAGE_PATH: str(self.packages_directory) or ""})

        self.hconfig_plugin_paths = self.extract_plugin_paths_from_HOUDINI_PATH(self.env_vars["HOUDINI_PATH"])

        files = next(os.walk(self.packages_directory))
        files = [name for name in files[2] if ".json" in name]  # only .json files

        # create each Package object
        for file in files:
            filename = Path(file).stem
            self.pkgs[filename] = Package(Path(self.packages_directory, file), self.hconfig_plugin_paths, self.env_vars)

    def extract_plugin_paths_from_HOUDINI_PATH(self, houdini_path: str) -> list[Path]:
        """
        Extract all the paths from the value of the HOUDINI_PATH environment variable.
        The paths are the directories that Houdini will search for the plugin data (HDAs/OTLs).

        Returns a list of pathlib.Path paths that exist.
        """

        plugin_paths = houdini_path.split(";")
        plugin_paths = [Path(path) for path in plugin_paths]
        plugin_paths = [path for path in plugin_paths if path.exists()]

        return plugin_paths


class Package:
    def __init__(
        self, config_path: Path, hconfig_plugin_paths: list[Path] = None, env_vars: dict[str, str] = None
    ) -> None:
        """
        A single JSON package file and its configuration and related data.

        If hconfig_plugin_paths is not given, manually determine which plugins the package are pointing to since
        we cannot use all_plugin_paths as a helping guide which was extracted from hconfig.

        Arguments:
            config_path (pathlib.Path):
                The file path of the package .json file.

            hconfig_plugin_paths (List[str]):
                The list of paths extracted from the HOUDINI_PATH environment variable produced by hconfig.
                These paths are associated with all packages for an installed version of Houdini.

            env_vars (Dict[str, str]):
                The environment variables that apply to all the packages for an
                installed Houdini version.
        """

        if not isinstance(config_path, Path):
            raise TypeError("package_path must be a pathlib.Path object.")

        if not hconfig_plugin_paths:
            hconfig_plugin_paths = []
        if not isinstance(hconfig_plugin_paths, list) or not all(
            isinstance(path, Path) for path in hconfig_plugin_paths
        ):
            raise TypeError("all_plugin_paths must be a list of pathlib.Path objects.")

        if env_vars and not isinstance(env_vars, dict):
            raise TypeError("env_vars must be a dict.")

        self.config_path = config_path

        self._hconfig_plugin_paths = hconfig_plugin_paths
        self._env_vars = env_vars or {}
        self._plugin_paths = []
        self.warnings = []

        self._load()
        self.config = self._flatten_package(self.config)

        self.resolve()
        self.extract_data()

        # initialize git repo data structure
        path = self.plugin_paths[0] if len(self.plugin_paths) != 0 else None
        self._git_project = GitProject(path)

    @property
    def env_vars(self) -> dict:
        """
        The environment variables that apply to all the packages for an installed Houdini version.
        """
        return self._env_vars

    ##############################################
    # set data determined by package's config file
    ##############################################

    @property
    def enable(self) -> bool:
        """
        Whether the package is enabled to be loaded by houdini.
        Can be found in package config.
        """
        enabled = self.is_enabled()
        return enabled

    @enable.setter
    def enable(self, toggle: bool) -> None:
        """
        Set the 'enable' variable's state of a plugin's config file to True or False.
        This determines if houdini loads the plugin.
        """

        if not isinstance(toggle, bool):
            raise TypeError("enable must be a bool.")

        self._enable = toggle

        # set in self.config
        if "enable" in self.config_keys:  # set existing key's value
            i = self.config_keys.index("enable")
            self.config[i][-1] = toggle
        else:  # create new key
            self.config.insert(0, ["enable", toggle])

        # set in .json config
        self._raw_json["enable"] = toggle  # enable is a top-level key, so this just works
        with open(self.config_path, "w") as outfile:
            json.dump(self._raw_json, outfile, indent=4)

    @property
    def pkg_name(self) -> str:
        """
        Name of the plugin's package (taken from the config's filename).
        """
        return self.config_path.stem

    @property
    def pkg_author(self) -> str:
        """
        The author of the package.
        """
        return self._git_project.owner

    @property
    def version_latest(self) -> str:
        """
        Latest version of the plugin (git tag version string).
        This is fetched from the remote repo via the GitHub API or a local json file if it already exists.
        """
        return self._git_project.remote.tag_latest

    @property
    def version_installed(self) -> str:
        """
        Installed version of the plugin (git tag version string).
        This is taken from local repo (if a repo exists).
        """
        return self._git_project.local.tag_latest

    @property
    def remote_repo_url(self) -> Url:
        """
        The URL of the plugin's remote repository, if it exists.
        """
        return self._git_project.remote_url

    @property
    def plugin_paths(self) -> list[Path]:
        """
        List of Path(s) to the local plugin data.
        Extracted from the package config file.
        """
        return self._plugin_paths

    @property
    def config_keys(self) -> list[str]:
        """
        A list of the keys of key-value pairs from a package's json config file.
        Convenient for checking the current saved state of a package configuration as recognized by houdini.
        """
        keys = [path[-2] for path in self.config]
        return keys

    @property
    def config_values(self) -> list[str | bool]:
        """
        A list of the values of key-value pairs from a package's json config file.
        Convenient for checking the current saved state of a package configuration as recognized by houdini.
        """
        values = [path[-1] for path in self.config]
        return values

    @property
    def last_metadata_sync_from_remote(self) -> str:
        """
        The date when this package's remote metadata was last synced from the remote repo.
        """
        pass

    @property
    def table_model(self) -> dict:
        """
        The model of package data that is later interpreted and loaded into the table GUI.
        The order of key-value pairs here determines how the columns are ordered in the GUI.
        The dict keys determine the names of the columns and which cells get which data.
        The dict values determine the contents of the row cells (that can be interpreted later by the table loader).
        """
        return {
            # first column in table is the index, this is created in the table view
            TableHeaders.ENABLE.value: self.enable,
            TableHeaders.PACKAGE.value: self.pkg_name,
            TableHeaders.AUTHOR.value: self.pkg_author,
            TableHeaders.INSTALLED.value: self.version_installed,
            TableHeaders.LATEST.value: self.version_latest,
            TableHeaders.SOURCE.value: self.remote_repo_url,
            TableHeaders.SYNC.value: None,  # functionality handled by table view
            TableHeaders.CONFIG.value: self.config_path,
            TableHeaders.PLUGINS.value: self.plugin_paths,
        }

    def is_enabled(self) -> bool:
        """
        Returns True or False based on whether or not the package is
        enabled according to its top-level "enable" variable.
        """

        if "enable" in self.config_keys:
            enabled = self.config_values[self.config_keys.index("enable")]
            # account for user using wrong type
            if isinstance(enabled, str) and enabled.lower() == "false":
                return False
            elif isinstance(enabled, str) and enabled.lower() == "true":
                return True
            return enabled
        return True

    def resolve(self) -> None:
        """
        Call all the methods necessary to convert the loaded package data into an easily
        readable format by flattening it and resolving all possible variables.

        Also tries to find which JSON packages are associated with which plugin paths.
        Since both are obtained separately (a matter of path of least resistance),
        they need to be associated with eachother. This is done by reading the
        package configs, extracting any paths within, and comparing them to the
        existing plugin paths.
        """

        config = self.config
        # use the global env vars to help resolve any variables in the package config
        env_vars = [list(item) for item in self._env_vars.items()]
        merged_config = env_vars + config  # prepend environment variables
        merged_config = self._resolve_vars(merged_config)
        # only get the original package config that is now variable-resolved.
        # no need to do anything: the original list is automatically updated since lists are mutable.
        self.config = config

    def extract_data(self) -> None:
        """
        Extract the relevant config data needed to populate the package manager Qt table.

        This operation can fail to get the data if the package has problems that the user needs to fix,
        such as circular variable references that can't be resolved. If a path contains a circular
        reference then this method will ignore it since it won't be a valid path.
        """

        plugin_paths_from_config = self._find_plugin_paths(self.config)

        # now compare the manually extracted plugin paths to the ones produced by hconfig
        # in order to find which plugin directories match which package files.
        # plugin paths and packages are obtained separately because that
        # is the easiest method of getting them.
        for path in plugin_paths_from_config:
            if path in self._hconfig_plugin_paths:
                self._plugin_paths.append(path)

    def _load(self) -> None:
        """
        Load json contents.
        Handles invalid json. If the json cannot be recovered, an empty dict will be set.
        """

        if not isinstance(self.config_path, Path):
            raise TypeError("path must be a pathlib.Path object.")

        class JSONPathDecoder(json.JSONDecoder):
            """
            Tries to parse invalid json into valid json by accounting for some possible errors.
            """

            def decode(self, s, **kwargs):
                regex_replacements = [
                    (re.compile(r"([^\\])\\([^\\])"), r"\1\\\\\2"),  # Fix single backslashes in paths
                    (re.compile(r",(\s*[\]}])"), r"\1"),  # Remove extraneous commas at the end of objects and arrays
                    (re.compile(r"}\s*{"), r"}, {"),  # Fix missing commas between objects
                ]
                for regex, replacement in regex_replacements:
                    s = regex.sub(replacement, s)
                return super().decode(s, **kwargs)

        try:
            with open(self.config_path) as f:
                data = json.load(f)
        except json.decoder.JSONDecodeError:
            logging.warning(f"Invalid json (might fail to resolve/parse): {self.config_path}")
            self.warnings.append("Invalid JSON! Fix errors and refresh this table.")
            try:
                with open(self.config_path) as f:
                    data = json.load(f, cls=JSONPathDecoder)
            except Exception:  # final catch all for all other broken json errors
                data = {}

        self._raw_json = data
        self.config = data

    def _flatten_package(self, data, prefix=None) -> list:
        """
        Recursively traverses a JSON-like data structure and returns a list of paths
        to each value. Each path is structured as its own list where each element is
        a key or the final value.

        Essentially produces a rudimentary tree data structure.

        Args:
            data (dict or list or scalar): A JSON-like data structure to traverse.
            prefix (list, optional): A list representing the current path being
                traversed. Defaults to None, in which case the prefix is initialized
                as an empty list.

        Returns:
            list: A list of paths, where each path is a list of keys and/or values
                representing a path to a value in the original data structure.
        """

        if prefix is None:
            prefix = []

        flat = []
        if isinstance(data, dict):
            for key, value in data.items():
                path = [*prefix, key]
                flat.extend(self._flatten_package(value, path))
        elif isinstance(data, list):
            for i, item in enumerate(data):
                path = [*prefix, i]
                flat.extend(self._flatten_package(item, path))
        else:  # str, int, bool
            flat.append([*prefix, data])
        return flat

    def _standard_paths(self, data: list[list]) -> list[list]:
        """
        Replace invalid double backslashes in paths with valid forward slashes.
        This ensures future regex operations do not encounter errors parsing escape characters.
        """

        for i, value in enumerate(data):
            value = value[-1]
            if isinstance(value, str) and ("/" in value or "\\" in value):
                value = value.replace("\\", "/")
                data[i][-1] = value
        return data

    def _split_indexes(self, nums: list[int], split_num: int) -> list[list[int]]:
        index = len(nums)
        for i, num in enumerate(nums):
            if num > split_num:
                index = i
                break

        start = nums[:index]
        end = nums[index:]
        return [start, end]

    def _replace_var_calls(self, data: list[list], var_calls: list, potential_var_names: list) -> list[list]:
        """
        Continuously replace variable calls with their respective values until no variable calls remain.
        Only replaces var calls if the variable exists to replace it with.
        Catches circular referencing variable calls.
        """

        # get all var inits
        # structure: [var name, var value, index of var initialization]
        var_inits = []
        known_vars = []
        [known_vars.append(call[0]) for call in var_calls if call[0] not in known_vars]
        for i, path in enumerate(data):
            if isinstance(path[-2], str) and path[-2].lower() in known_vars:
                var_inits.append([path[-2], path[-1], i])

        # replace var calls with var values
        for call, call_i, processed_vars in var_calls:
            # check for circular references
            if call in processed_vars:
                self.warnings.append(f"Can't process package! Circular reference detected for variable: '{call}'")
                return
            processed_vars.add(call)

            # determine which var init to try to get value from first
            # get value from var init closest to var call and before the var call, or after
            var_init_indexes = [init[2] for init in var_inits if init[0].lower() == call]
            var_init_priority = self._split_indexes(var_init_indexes, call_i)
            var_index = var_init_priority[0][-1] if len(var_init_priority[0]) != 0 else var_init_priority[1][0]

            var_init_indexes = [init[2] for init in var_inits]
            var = var_inits[var_init_indexes.index(var_index)]

            # case insensitive replace
            compiled = re.compile(re.escape("$" + call), re.IGNORECASE)
            data[call_i][-1] = compiled.sub(var[1], data[call_i][-1])

            # check if the new value contains variable calls
            new_value = data[call_i][-1]
            if "$" in new_value:
                new_var_calls = [
                    [call, call_i, processed_vars.copy()]
                    for call in (
                        "".join(takewhile(str.isidentifier, call.lower())) for call in new_value.split("$")[1:]
                    )
                    if call and call in potential_var_names
                ]
                self._replace_var_calls(data, new_var_calls, potential_var_names)

        return data

    def _resolve_vars(self, config: list[list]) -> None:
        """
        Replace every variable call in a package config with the variable's value.

        Custom package data structure rules:
            The last list element is a value.
            The second last list element is a value's key.
            When variable call found, work backwards from the call to find the most recent
            declaration of variable (var names will always be the second last element).

        Arguments:
            config (list):
                A flattened package config. Each item of each sublist is a component
                of a "path" that points to the last value.
                e.g. ['env', 0, 'HOUDINI_PATH', 'C:/Users/user/Desktop/myplugin']
        """

        def is_variable(char: str) -> bool:
            """
            Check if a string is a valid character that would make up a standard variable name.
            """

            return char.isalpha() or char.isdigit() or char == "_"

        config = self._standard_paths(config)

        while True:
            # get list of potential variables
            potential_var_names = []
            [
                potential_var_names.append(path[-2].lower())
                for path in config
                if isinstance(path[-2], str) and len(path) >= 2 and path[-2].lower() not in potential_var_names
            ]

            # find and extract variable calls
            var_calls = []
            for i, path in enumerate(config):
                if isinstance(path[-1], str) and "$" in path[-1]:
                    var_calls.extend(
                        [call, i, set()]
                        for call in ("".join(takewhile(is_variable, call.lower())) for call in path[-1].split("$")[1:])
                        if call and call in potential_var_names
                    )

            # break the loop if no more variable calls are found or there are errors with the package that can't be parsed
            if not var_calls or self.warnings:
                break

            config = self._replace_var_calls(config, var_calls, potential_var_names)

        return config

    def _find_plugin_paths(self, paths: list[list]) -> list[str]:
        """
        Find all the plugin paths in the package.
        Returns a list of all the valid paths.
        """

        def split_paths(string: str) -> list:
            """
            Split a combined string of multiple paths into individual paths.
            Removes trailing separators.
            """
            if string[-2:] == ";&":
                string = string[:-2]
            elif string[-1] == ";":
                string = string[:-1]
            string = string.split(";")
            return string

        # locate HOUDINI_PATH
        # Need to look for "path" (legacy of HOUDINI_PATH) as well since our manual reading of
        # the package config does not cause any paths in "path" to be automatically merged into HOUDINI_PATH,
        # as apposed to when packages are read by hconfig
        # HOUDINI_PATH or "path" can be anywhere in the chain, not only just [-2]
        paths = [path for path in paths if "HOUDINI_PATH" in path or "path" in path or "hpath" in path]
        paths = [path for path in paths if isinstance(path[-1], str)]
        new_paths = []
        [new_paths.extend(split_paths(path[-1])) for path in paths]
        paths = new_paths

        # remove duplicate paths
        new_paths = []
        [new_paths.extend([path]) for path in paths if path not in new_paths]
        paths = new_paths

        paths = [Path(path) for path in paths]
        paths = [path for path in paths if path.exists()]
        return paths
